import datetime
import subprocess
import pathlib

import sqlalchemy as sa
from sqlalchemy import create_engine

from acqdiv.database.model import Base
import acqdiv.database.model as db
from acqdiv.util.path import get_full_path


class DBProcessor:
    """Methods for adding corpus data to the database."""

    def __init__(self, db_dir='database'):
        """Initialize DB engine.

        Args:
            db_dir (str): Where the database is written to.
        """
        self.engine = self.get_engine(db_dir)

        # initialize them once for each session
        # to increase performance
        self.insert_corpus_func = None
        self.insert_session_func = None
        self.insert_speaker_func = None
        self.insert_utt_func = None
        self.insert_word_func = None
        self.insert_morph_func = None

    @classmethod
    def get_engine(cls, db_dir):
        """Return a database engine.

        Args:
            db_dir (str): Where the database is written to.

        Returns:
            Engine: The DB engine.
        """
        db_dir = pathlib.Path(db_dir)
        date = datetime.datetime.now().strftime('%Y-%m-%d')
        path = db_dir / f'acqdiv_corpus_{date}.sqlite3'

        print(f"Writing database to: {path.resolve()}")
        print()
        engine = create_engine(f'sqlite:///{str(path)}', echo=False)
        cls.create_tables(engine)
        cls.create_views(path)

        return engine

    @staticmethod
    def create_tables(engine):
        """Drop all tables before creating them.

            Args:
                engine: An sqlalchemy database engine.
        """
        Base.metadata.drop_all(bind=engine)
        Base.metadata.create_all(engine)

    @staticmethod
    def create_views(db_path):
        view_path = get_full_path('database/views.sql')
        cmd = f'sqlite3 {db_path} < {view_path}'
        subprocess.run(cmd, shell=True)

    def insert_corpus(self, corpus):
        """Insert the corpus into the database.

        Args:
            corpus (acqdiv.model.corpus.Corpus): The corpus.
        """
        with self.engine.begin() as conn:
            conn.execute('PRAGMA synchronous = OFF')
            conn.execute('PRAGMA journal_mode = MEMORY')
            conn = conn.execution_options(compiled_cache={})
            self.insert_corpus_func = sa.insert(db.Corpus, bind=conn).execute
            c_id = self.insert_corpus_metadata(corpus)

        uspeakers_dict = {}

        for session in corpus.sessions:
            self.insert_session(session, c_id, uspeakers_dict)

    def insert_corpus_metadata(self, corpus):
        """Insert the data into the `Corpus` table.

        Args:
            corpus (acqdiv.model.corpus.Corpus): The corpus.

        Returns:
            str: The ID of the corpus.
        """
        c_id, = self.insert_corpus_func(
            id=corpus.corpus,
            language=corpus.language,
            iso_639_3=corpus.iso_639_3,
            glottolog_code=corpus.glottolog_code,
            owner=corpus.owner,
            acronym=corpus.acronym,
            name=corpus.name,
        ).inserted_primary_key

        return c_id

    def insert_session(self, session, c_id, uspeakers_dict):
        """Insert the session into the database.

        Args:
            session (acqdiv.model.session.Session): The session.
            c_id (str): The corpus ID.
        """
        with self.engine.begin() as conn:
            conn.execute('PRAGMA synchronous = OFF')
            conn.execute('PRAGMA journal_mode = MEMORY')
            conn = conn.execution_options(compiled_cache={})

            self.insert_session_func = sa.insert(db.Session, bind=conn).execute
            self.insert_speaker_func = sa.insert(db.Speaker, bind=conn).execute
            self.insert_uspeaker_func = sa.insert(
                db.UniqueSpeaker, bind=conn).execute
            self.insert_utt_func = sa.insert(db.Utterance, bind=conn).execute
            self.insert_word_func = sa.insert(db.Word, bind=conn).execute
            self.insert_morph_func = sa.insert(db.Morpheme, bind=conn).execute

            s_id = self.insert_session_metadata(session, c_id)
            speakers_dict = self.insert_speakers(
                session.speakers, s_id, c_id, uspeakers_dict)
            self.insert_utterances(session.utterances, s_id, speakers_dict)

    def insert_session_metadata(self, session, c_id):
        s_id, = self.insert_session_func(
            corpus=c_id,
            date=session.date,
            source_id=session.source_id,
            duration=session.duration if session.duration else None,
            media_id=session.media_filename if session.media_filename else None
        ).inserted_primary_key

        return s_id

    def insert_speakers(self, speakers, s_id, c_id, uspeakers_dict):
        speakers_dict = {
            None: None
        }
        for speaker in speakers:
            if speaker.uniquespeaker in uspeakers_dict:
                usp_id = uspeakers_dict[speaker.uniquespeaker]
            else:
                usp_id = self.insert_uspeaker(speaker.uniquespeaker, c_id)
                uspeakers_dict[speaker.uniquespeaker] = usp_id

            sp_id = self.insert_speaker(speaker, s_id, usp_id)
            speakers_dict[speaker] = sp_id

        return speakers_dict

    def insert_uspeaker(self, uspeaker, c_id):
        usp_id, = self.insert_uspeaker_func(
            corpus=c_id,
            name=uspeaker.name if uspeaker.name else None,
            birthdate=uspeaker.birth_date if uspeaker.birth_date else None,
            gender_raw=uspeaker.gender_raw if uspeaker.gender_raw else None,
            gender=uspeaker.gender if uspeaker.gender else None,
            speaker_label=uspeaker.code if uspeaker.code else None,
        ).inserted_primary_key

        return usp_id

    def insert_speaker(self, speaker, s_id, usp_id):
        sp_id, = self.insert_speaker_func(
            session_id_fk=s_id,
            uniquespeaker_id_fk=usp_id,
            age_raw=speaker.age_raw if speaker.age_raw else None,
            age=speaker.age if speaker.age else None,
            age_in_days=speaker.age_in_days if speaker.age_in_days else None,
            role_raw=speaker.role_raw if speaker.role_raw else None,
            role=speaker.role if speaker.role else None,
            macrorole=speaker.macro_role if speaker.macro_role else None,
            languages_spoken=speaker.languages_spoken
            if speaker.languages_spoken else None
        ).inserted_primary_key

        return sp_id

    def insert_utterances(self, utterances, s_id, speakers_dict):
        for utt in utterances:
            u_id = self.insert_utterance(utt, s_id, speakers_dict)
            w_ids = self.insert_words(utt.words, u_id)
            self.insert_morphemes(utt.morphemes, u_id, w_ids)

    def insert_utterance(self, utt, s_id, speakers_dict):
        u_id, = self.insert_utt_func(
            session_id_fk=s_id,
            source_id=utt.source_id,
            speaker_id_fk=speakers_dict[utt.speaker],
            addressee_id_fk=speakers_dict[utt.addressee],
            utterance_raw=utt.utterance_raw if utt.utterance_raw else None,
            utterance=utt.utterance if utt.utterance else None,
            translation=utt.translation if utt.translation else None,
            morpheme=utt.morpheme_raw if utt.morpheme_raw else None,
            gloss_raw=utt.gloss_raw if utt.gloss_raw else None,
            pos_raw=utt.pos_raw if utt.pos_raw else None,
            sentence_type=utt.sentence_type if utt.sentence_type else None,
            childdirected=utt.childdirected
            if isinstance(utt.childdirected, bool) else None,
            start_raw=utt.start_raw if utt.start_raw else None,
            start=utt.start if utt.start else None,
            end_raw=utt.end_raw if utt.end_raw else None,
            end=utt.end if utt.end else None,
            comment=utt.comment if utt.comment else None,
        ).inserted_primary_key

        return u_id

    def insert_words(self, words, u_id):
        w_ids = []
        for w in words:
            w_id = self.insert_word(w, u_id)
            w_ids.append(w_id)

        return w_ids

    def insert_word(self, w, u_id):
        w_id, = self.insert_word_func(
            utterance_id_fk=u_id,
            language=w.word_language if w.word_language else None,
            word=w.word if w.word else None,
            word_actual=w.word_actual if w.word_actual else None,
            word_target=w.word_target if w.word_target else None,
            pos=w.pos if w.pos else None,
            pos_ud=w.pos_ud if w.pos_ud else None,

        ).inserted_primary_key

        return w_id

    def insert_morphemes(self, morphemes, u_id, w_ids):
        link_to_word = len(morphemes) == len(w_ids)

        for i, mword in enumerate(morphemes):
            w_id = w_ids[i] if link_to_word else None

            for m in mword:
                self.insert_morpheme(m, u_id, w_id)

    def insert_morpheme(self, m, u_id, w_id):
        """Insert the morpheme.

        Args:
            m (acqdiv.model.morpheme.Morpheme): The morpheme instance.
            u_id (str): The utterance ID.
            w_id (str): The word ID.
        """
        self.insert_morph_func(
            utterance_id_fk=u_id,
            word_id_fk=w_id,
            language=m.morpheme_language if m.morpheme_language else None,
            type=m.type if m.type else None,
            morpheme=m.morpheme if m.morpheme else None,
            gloss_raw=m.gloss_raw if m.gloss_raw else None,
            gloss=m.gloss if m.gloss else None,
            pos_raw=m.pos_raw if m.pos_raw else None,
            pos=m.pos if m.pos else None,
            lemma_id=m.lemma_id if m.lemma_id else None,
        )



import re

from acqdiv.parsers.metadata.imdi_parser import IMDIParser


class QaqetIMDI(IMDIParser):

    def get_tc_code(self, session_code):
        return re.search(r'[A-Z]{3}', session_code).group()

    def get_participants(self):
        participants = super().get_participants()
        tc_code = self.get_tc_code(self.root.Session.Name.text)

        for participant in participants:
            if participant['code'] == tc_code:
                participant['role'] = 'Target_Child'
            else:
                participant['role'] = participant['familysocialrole']

        return participants



import re


class ActualTargetUtteranceExtractor:
    """Methods for extracting actual and target utterances."""

    @classmethod
    def to_actual_utterance(cls, utterance):
        """Extract actual utterance."""
        for actual_method in [cls.get_shortening_actual,
                              cls.get_fragment_actual,
                              cls.get_replacement_actual]:
            utterance = actual_method(utterance)

        return utterance

    @classmethod
    def to_target_utterance(cls, utterance):
        """Extract target utterance."""
        for target_method in [cls.get_shortening_target,
                              cls.get_fragment_target,
                              cls.get_replacement_target]:
            utterance = target_method(utterance)

        return utterance

    @staticmethod
    def get_shortening_actual(utterance):
        """Get the actual form of shortenings.

        Coding in CHAT: parentheses within word.
        The part with parentheses is removed.
        """
        shortening_regex = re.compile(r'(?<=\S)\(\S+?\)|\(\S+?\)(?=\S)')
        return shortening_regex.sub('', utterance)

    @staticmethod
    def get_shortening_target(utterance):
        """Get the target form of shortenings.

        Coding in CHAT: parentheses within word.
        The part in parentheses is kept, parentheses are removed.
        """
        shortening_regex = re.compile(r'(?<=\S)\((\S+?)\)|\((\S+?)\)(?=\S)')
        return shortening_regex.sub(r'\1\2', utterance)

    @staticmethod
    def get_replacement_actual(utterance):
        """Get the actual form of replacements.

        Coding in CHAT: [: <words>] .
        Keeps replaced words, removes replacing words with brackets.
        """
        # several scoped words
        replacement_regex1 = re.compile(r'<(.*?)> ?\[: .*?\]')
        clean = replacement_regex1.sub(r'\1', utterance)
        # one scoped word
        replacement_regex2 = re.compile(r'(\S+) ?\[: .*?\]')
        return replacement_regex2.sub(r'\1', clean)

    @staticmethod
    def get_replacement_target(utterance):
        """Get the target form of replacements.

        Coding in CHAT: [: <words>] .
        Removes replaced words, keeps replacing words within brackets. If there
        is more than one replacing word, they are joined together by an
        underscore.
        """
        replacement_regex = re.compile(r'(?:<.*?>|\S+) ?\[: (.*?)\]')

        def x(match):
            return match.group(1).replace(' ', '_')

        return replacement_regex.sub(x, utterance)

    @staticmethod
    def get_fragment_actual(utterance):
        """Get the actual form of fragments.

        Coding in CHAT: word starting with &.
        Keeps the fragment, removes the & from the word.
        """
        fragment_regex = re.compile(r'(^|\s)&([^-=\s]\S*)')
        return fragment_regex.sub(r'\1\2', utterance)

    @staticmethod
    def get_fragment_target(utterance):
        """Get the target form of fragments.

        Coding in CHAT: word starting with &.
        The fragment is marked as untranscribed (xxx).
        """
        fragment_regex = re.compile(r'(^|\s)&([^-=\s]\S*)')
        return fragment_regex.sub(r'\1xxx', utterance)

    @staticmethod
    def get_retracing_actual(utterance):
        """Get the actual form of retracings.

        Coding in CHAT: [/], [//], [///], [/-]

        Removal of retracing markers.
        """
        # several scoped words
        retracing_regex1 = re.compile(r'<(.*?)> ?\[(/{1,3}|/-)\]')
        clean = retracing_regex1.sub(r'\1', utterance)
        # one scoped word
        retracing_regex2 = re.compile(r'(\S+) ?\[(/{1,3}|/-)\]')
        return retracing_regex2.sub(r'\1', clean)

    @classmethod
    def get_retracing_target(cls, utterance):
        """Get the target form of retracings.

        Coding in CHAT: [/], [//], [///], [/-]

        Same treatment as the actual form, except for single-word
        corrections in which both the retraced and retracing part receive
        the retracing value, e.g. hui [//] hoi du -> hoi hoi du. Multiple words
        being corrected cannot be replaced by their target forms since the
        correcting part can be of variable length.
        """
        # single-word correction
        retracing_regex = re.compile(r'([^>\s]+) ?\[//\] (\S+)')
        utterance = retracing_regex.sub(r'\2 \2', utterance)
        return cls.get_retracing_actual(utterance)



from acqdiv.parsers.toolbox.cleaners.morpheme_cleaner \
    import ToolboxMorphemeCleaner
from acqdiv.util.csvparser import parse_csv
from acqdiv.util.path import get_full_path


class IndonesianGlossMapper:

    gloss_dict = parse_csv(get_full_path(
        'parsers/corpora/main/indonesian/resources/gloss.csv'))

    @classmethod
    def map(cls, gloss):
        gloss = ToolboxMorphemeCleaner.remove_morpheme_delimiters(gloss)
        return cls.gloss_dict.get(gloss, '')



from acqdiv.parsers.toolbox.readers.reader import ToolboxReader


class TuatschinReader(ToolboxReader):

    @staticmethod
    def get_rec_separator():
        return br'\\u_id'

    @classmethod
    def get_source_id(cls, rec):
        return rec.get('u_id')

    @classmethod
    def get_actual_utterance(cls, rec):
        """Get the raw utterance.

        \token is the normalized/tokenized tier of \text which is the very
        original transcription. \token is taken as the raw utterance.
        """
        return rec.get('token', '')

    @classmethod
    def get_sentence_type(cls, rec):
        """Get sentence type.

        There are only explicit question ('?'), exclamation ('!') and trail
        off ('. . .') markers. If they are missing, declarative (default)
        sentence type is assumed.
        """
        utterance = cls.get_actual_utterance(rec)
        # remove the typical trailing whitespaces
        utterance = utterance.rstrip()
        if utterance.endswith('?'):
            return 'question'
        elif utterance.endswith('!'):
            return 'exclamation'
        elif utterance.endswith('. . .'):
            return 'trail off'
        else:
            return 'default'

    @classmethod
    def get_translation(cls, rec):
        """Get translation.

        There is no free english translation, only a lemmatized translation
        is available.
        """
        return rec.get('lemma_en', '')

    @classmethod
    def get_comment(cls, rec):
        return rec.get('com', '')

    @classmethod
    def get_seg_tier(cls, rec):
        """Get segment tier.

        No segment, but only lemmas are available.
        """
        return rec.get('lemma', '')

    @classmethod
    def get_gloss_tier(cls, rec):
        """Get the gloss tier.

        Only morphsyntactical glosses (no lexical glosses) are available.
        """
        return rec.get('morphosyn', '')

    @classmethod
    def get_pos_tier(cls, rec):
        return rec.get('pos', '')

    @classmethod
    def get_morphemes(cls, morpheme_word):
        """Get morphemes.

        There is no segmentation of words into morphemes.
        """
        return [morpheme_word]

    @staticmethod
    def get_morpheme_type():
        return 'actual'



import re

from acqdiv.util.csvparser import parse_csv, parse_pos_ud
from acqdiv.util.path import get_full_path


class TuatschinPOSMapper:

    pos_dict = parse_csv(get_full_path(
        'parsers/corpora/main/tuatschin/resources/pos.csv'))

    pos_ud_dict = parse_pos_ud(get_full_path(
        'parsers/corpora/main/tuatschin/resources/pos.csv'))

    @classmethod
    def map(cls, pos, ud=False):
        pos = cls.clean_pos(pos)

        if ud:
            return cls.pos_ud_dict.get(pos, '')
        else:
            return cls.pos_dict.get(pos, '')

    @classmethod
    def clean_pos(cls, pos):
        for cleaning_method in [
            cls.remove_specifications
        ]:
            pos = cleaning_method(pos)
        return pos

    @staticmethod
    def remove_specifications(pos):
        """Remove specifications of POS tags.

        Specifications start with `_`.

        Examples:
        - words erroneously written apart: _cont
        - child forms: _Chld
        - discourse particles: _Discpart
        ...
        """
        regex = re.compile(r'_[^_]+')
        pos = regex.sub('', pos)
        return pos



def infer_childdirected(utt):
    """Infer child directedness.

    Args:
        utt (acqdiv.model.utterance.Utterance): The utterance.
    """
    if utt.childdirected == '':
        if utt.addressee:
            if (utt.addressee.macro_role == 'Target_Child'
                    and utt.addressee != utt.speaker):
                return True
            else:
                return False

    return utt.childdirected



import re

from acqdiv.parsers.chat.readers.fileparser import CHATFileParser
from acqdiv.parsers.chat.readers.actual_target_utterance \
    import ActualTargetUtteranceExtractor
from acqdiv.parsers.chat.readers.sentence_type \
    import SentenceTypeExtractor


class CHATReader:
    """Methods for reading CHAT files."""

    def __init__(self, session_file):
        """Set the variables.
        chat (acqdiv.parsers.chat.model.CHAT): The chat instance.
        participant (acqdiv.parsers.chat.model.Participant): The current
            Participant
        record (acqdiv.parsers.chat.model.Record): The current record.
        """
        self.chat = CHATFileParser.parse(session_file)

        self.participant_iterator = iter(self.chat.participants)
        self.participant = None

        self.record_iterator = iter(self.chat.records)
        self.record = None

    # ---------- metadata ----------

    def get_session_date(self):
        """Get the date of the session.

        Returns: str
        """
        return self.chat.date

    def get_session_media_filename(self):
        """Get the the media file name of the session.

        Returns: str
        """
        return self.chat.media_filename

    def get_target_child(self):
        """Get the target child of the session.

        Returns:
            Tuple[str, str]: (label, name)
        """
        for code, p in self.chat.participants.items():
            if p.role == 'Target_Child':
                return code, p.name

        return '', ''

    # ---------- speaker ----------

    def load_next_speaker(self):
        """Load the data for the next speaker.

        Returns:
            bool: 1 if a new speaker could be loaded, otherwise 0.
        """
        try:
            code = next(self.participant_iterator)
            self.participant = self.chat.participants[code]
        except StopIteration:
            return 0

        return 1

    def get_speaker_age(self):
        """Get the age of the speaker.

        Returns: str
        """
        return self.participant.age

    def get_speaker_birthdate(self):
        """Get the birth date of the speaker.

        Returns: str
        """
        return self.participant.birth_date

    def get_speaker_gender(self):
        """Get the gender of the speaker.

        Returns: str
        """
        return self.participant.sex

    def get_speaker_label(self):
        """Get the label of the speaker.

        Returns: str
        """
        return self.participant.code

    def get_speaker_language(self):
        """Get the language of the speaker.

        Returns: str
        """
        return self.participant.language

    def get_speaker_name(self):
        """Get the name of the speaker.

        Returns: str
        """
        return self.participant.name

    def get_speaker_role(self):
        """Get the role of the speaker.

        Returns: str
        """
        return self.participant.role

    # ---------- record ----------

    def load_next_record(self):
        """Load the next record.

        Returns:
            bool: 1 if a new record could be loaded, otherwise 0.
        """
        try:
            self.record = next(self.record_iterator)
        except StopIteration:
            return 0

        return 1

    def get_uid(self):
        """Get the ID of the utterance.

        Returns: str
        """
        return str(self.record.uid)

    # ---------- main line ----------

    def get_utterance(self):
        """Get the utterance.

        Returns: str
        """
        return self.record.utterance

    def get_sentence_type(self):
        """Get the sentence type of the utterance.

        Returns: str
        """
        utterance = self.get_utterance()
        return SentenceTypeExtractor.get_sentence_type(utterance)

    def get_actual_utterance(self):
        """Get the actual form of the utterance.

        Returns: str
        """
        utterance = self.get_utterance()
        return ActualTargetUtteranceExtractor.to_actual_utterance(utterance)

    def get_target_utterance(self):
        """Get the target form of the utterance.

        Returns: str
        """
        utterance = self.get_utterance()
        return ActualTargetUtteranceExtractor.to_target_utterance(utterance)

    def get_record_speaker_label(self):
        """Get the label of the speaker of the utterance.

        Returns: str
        """
        return self.record.participant_code

    def get_start_time(self):
        """Get the start time of the utterance.

        Returns: str
        """
        return self.record.start_time

    def get_end_time(self):
        """Get the end time of the utterance.

        Returns: str
        """
        return self.record.end_time

    # ---------- dependent tiers ----------

    def get_addressee(self):
        """Get the addressee of the utterance.

        Returns: str
        """
        return self.record.dependent_tiers.get('add', '')

    def get_translation(self):
        """Get the translation of the utterance.

        Returns: str
        """
        return self.record.dependent_tiers.get('eng', '')

    def get_comments(self):
        """Get the comments of the utterance.

        Returns: str
        """
        comments = self.record.dependent_tiers.get('com', '')
        situation = self.record.dependent_tiers.get('sit', '')
        action = self.record.dependent_tiers.get('act', '')
        explanation = self.record.dependent_tiers.get('exp', '')
        fields = [comments, situation, action, explanation]

        return '; '.join((f for f in fields if f))

    def get_morph_tier(self):
        """Get the morphology tier.

        Per default, this is the 'mor'-tier.

        Returns:
            str: The content of the morphology tier.
        """
        return self.record.dependent_tiers.get('mor', '')

    def get_seg_tier(self):
        """Get the tier containing segments.

        Returns: str
        """
        return self.get_morph_tier()

    def get_gloss_tier(self):
        """Get the tier containing glosses.

        Returns: str
        """
        return self.get_morph_tier()

    def get_pos_tier(self):
        """Get the tier containing POS tags.

        Returns: str
        """
        return self.get_morph_tier()

    # ---------- words ----------

    @staticmethod
    def get_utterance_words(utterance):
        """Get the words of an utterance.

        Words are defined as units separated by a blank space.

        Args:
            utterance (str): The utterance.

        Returns:
            list: The words.
        """
        if utterance:
            return re.split(r'\s+', utterance)
        else:
            return []

    @classmethod
    def get_morpheme_words(cls, morph_tier):
        return cls.get_utterance_words(morph_tier)

    @classmethod
    def get_seg_words(cls, seg_tier):
        """Get the words from the segment tier.

        Returns: list
        """
        return cls.get_morpheme_words(seg_tier)

    @classmethod
    def get_gloss_words(cls, gloss_tier):
        """Get the words from the gloss tier.

        Returns: list
        """
        return cls.get_morpheme_words(gloss_tier)

    @classmethod
    def get_pos_words(cls, pos_tier):
        """Get the words from the POS tag tier.

        Returns: list
        """
        return cls.get_morpheme_words(pos_tier)

    @staticmethod
    def get_word_language(word):
        """Get the language of the word.

        Returns: str
        """
        return ''

    @staticmethod
    def get_standard_form():
        """Get the standard form of the utterance.

        Returns:
            str: 'actual' or 'target'
        """
        return 'actual'

    # ---------- morphemes ----------

    @staticmethod
    def get_morphemes(morpheme_word):
        """Get morphemes of a word.

        Per default, morphemes are separated by a dash.

        Args:
            morpheme_word (str): Word consisting of morphemes.

        Returns:
            list: The morphemes of the word.
        """
        return morpheme_word.split('-')

    @classmethod
    def get_segments(cls, seg_word):
        """Get the segments from the segment word.

        Returns: list
        """
        return cls.get_morphemes(seg_word)

    @classmethod
    def get_glosses(cls, gloss_word):
        """Get the glosses from the gloss word.

        Returns: list
        """
        return cls.get_morphemes(gloss_word)

    @classmethod
    def get_poses(cls, pos_word):
        """Get the POS tags from the POS word.

        Returns: list
        """
        return cls.get_morphemes(pos_word)

    @staticmethod
    def get_morpheme_language(seg, gloss, pos):
        """Get language of the morpheme.

        Returns: str
        """
        return ''

    @staticmethod
    def get_main_morpheme():
        """Get the main morpheme.

        The main morpheme is the one used for linking the words of the
        respective morphology tier to those of the utterance. In case of
        misalignments between the morphology tiers, only the main morphemes
        are kept, while those of the other morphology tiers are nulled.

        Returns:
            str: 'segment' or 'gloss'.
        """
        return 'gloss'

    @staticmethod
    def get_morpheme_type():
        return 'target'



from acqdiv.util.csvparser import parse_csv, parse_pos_ud
from acqdiv.util.path import get_full_path


class NungonPOSMapper:

    pos_dict = parse_csv(get_full_path(
        'parsers/corpora/main/nungon/resources/pos.csv'))

    pos_ud_dict = parse_pos_ud(get_full_path(
        'parsers/corpora/main/nungon/resources/pos.csv'))

    @classmethod
    def map(cls, pos, ud=False):
        pos = cls.remove_question_mark(pos)

        if ud:
            return cls.pos_ud_dict.get(pos, '')
        else:
            return cls.pos_dict.get(pos, '')

    @staticmethod
    def remove_question_mark(morpheme):
        """Remove the question mark in the morpheme.

        Question marks might code insecure annotations. They are prefixed to
        the morpheme.
        """
        return morpheme.lstrip('?')



import re

from acqdiv.parsers.chat.readers.reader import CHATReader

from acqdiv.parsers.corpora.main.\
    japanese_miipro.gloss2segment_mapper \
    import JapaneseMiiProGloss2SegmentMapper as Mp


class JapaneseMiiProReader(CHATReader):

    @classmethod
    def _is_target_child(cls, pos, role):
        if role == 'Target_Child':
            return True

        if role == 'Child' and pos == 0:
            return True

        return False

    def get_target_child(self):
        for pos, (code, p) in enumerate(self.chat.participants.items()):
            if self._is_target_child(pos, p.role):
                return code, p.name

        return '', ''

    def get_morph_tier(self):
        if self.record.dependent_tiers.get('xtrn', ''):
            return self.record.dependent_tiers.get('xtrn', '')
        else:
            return self.record.dependent_tiers.get('xmor', '')

    @staticmethod
    def get_word_language(word):
        if word.endswith('@s:eng'):
            return 'English'
        else:
            return 'Japanese'

    @staticmethod
    def iter_morphemes(morph_word):
        """Iter morphemes of a word.

        A word consists of word groups in the case of compounds (marker: +).

        A word group has the following structure:
        prefix#POS|stem&fusionalsuffix-suffix=gloss

        prefix: segment, no gloss, no POS (-> assign 'pfx')
        stem: segment, gloss, POS
        suffix: no segment, with gloss, without POS (-> assign 'sfx')

        For every component of the compound '=' is prepended to the part (e.g.
        'n|+n|apple+n|tree' -> '=apple', '=tree'). Both parts receive the same
        gloss. The POS tag of the whole compound is removed. There may be
        prefixes attached to the whole compound. There are no clitics.

        Suffixes with a colon are specially treated: If the part after the
        colon is not 'contr' (contraction), it denotes the segment of the
        suffix.

        Returns:
            tuple: (segment, gloss, pos).
        """
        morpheme_regex = re.compile(r'[^#]+#'
                                    r'|[^\-]+'
                                    r'|[\-][^\-]+')

        # get stem gloss and remove it from morpheme word
        match = re.search(r'(.+)=(\S+)$', morph_word)
        if match:
            morph_word = match.group(1)
            stem_gloss = match.group(2)
        else:
            stem_gloss = ''

        # split into word groups (i.e. into compound parts) (if applicable)
        word_groups = morph_word.split('+')

        # check if word is a compound
        if len(word_groups) > 1:
            # pop prefixes & the POS tag of the whole compound
            pfxs_cmppos = word_groups.pop(0)

            # iter prefixes preceding compound
            for pfx_match in re.finditer(r'[^#]+(?=#)', pfxs_cmppos):
                yield pfx_match.group(), '', 'pfx'

        for word_group in word_groups:

            # iter morphemes
            for match in morpheme_regex.finditer(word_group):
                morpheme = match.group()

                # prefix
                if morpheme.endswith('#'):
                    segment = morpheme.rstrip('#')
                    gloss = ''
                    pos = 'pfx'
                # sfx
                elif morpheme.startswith('-'):
                    sfx = morpheme.lstrip('-')
                    pos = 'sfx'
                    match = re.search(r'([^:]+)(:(.*))?', sfx)
                    # check for colon case
                    if match.group(2) and match.group(3) != 'contr':
                        segment = match.group(3)
                        gloss = match.group(1)
                    else:
                        segment = ''
                        gloss = sfx
                # stem
                else:
                    pos, segment = morpheme.split('|')
                    # if it is a compound part
                    if len(word_groups) > 1:
                        # prepend '=' to segment
                        segment = '=' + segment
                    gloss = stem_gloss

                if Mp.map(gloss):
                    segment = Mp.map(gloss)

                yield segment, gloss, pos

    @classmethod
    def get_segments(cls, seg_word):
        return [seg for seg, _, _ in cls.iter_morphemes(seg_word)]

    @classmethod
    def get_glosses(cls, gloss_word):
        return [gloss for _, gloss, _ in cls.iter_morphemes(gloss_word)]

    @classmethod
    def get_poses(cls, pos_word):
        return [pos for _, _, pos in cls.iter_morphemes(pos_word)]



from acqdiv.parsers.chat.readers.reader import CHATReader


class CreeReader(CHATReader):

    @staticmethod
    def get_main_morpheme():
        return 'segment'

    def get_seg_tier(self):
        return self.record.dependent_tiers.get('xtarmor', '')

    def get_gloss_tier(self):
        return self.record.dependent_tiers.get('xmormea', '')

    def get_pos_tier(self):
        return self.record.dependent_tiers.get('xmortyp', '')

    @staticmethod
    def get_morphemes(word):
        if word:
            return word.split('~')
        else:
            return []

    @classmethod
    def get_segments(cls, seg_word):
        return cls.get_morphemes(seg_word)

    @classmethod
    def get_glosses(cls, gloss_word):
        return cls.get_morphemes(gloss_word)

    @classmethod
    def get_poses(cls, pos_word):
        return cls.get_morphemes(pos_word)

    @staticmethod
    def get_morpheme_language(seg, gloss, pos):
        if gloss == 'Eng':
            return 'English'
        else:
            return 'Cree'



from acqdiv.parsers.corpora.main.tuatschin.reader \
    import TuatschinReader
from acqdiv.parsers.corpora.main.tuatschin.cleaner \
    import TuatschinCleaner
from acqdiv.parsers.metadata.cmdi_parser import CMDIParser
from acqdiv.parsers.toolbox.parser import ToolboxParser


class TuatschinSessionParser(ToolboxParser):

    def get_metadata_reader(self):
        return CMDIParser(self.metadata_path)

    def get_record_reader(self):
        return TuatschinReader()

    def get_cleaner(self):
        return TuatschinCleaner()

    def add_record(self, rec):
        super().add_record(rec)
        self.delete_morphemes()

    def delete_morphemes(self):
        utt = self.session.utterances[-1]
        utt.morpheme_raw = ''
        utt.gloss_raw = ''
        utt.pos_raw = ''
        utt.morphemes = []



""" Entry point for loading ACQDIV raw input corpora data into the ACQDIV-DB
"""
import os
from configparser import ConfigParser, ExtendedInterpolation

from acqdiv.parsers.corpus_parser_mapper import CorpusParserMapper
from acqdiv.database.processor import DBProcessor


class Loader:

    @staticmethod
    def load(cfg_path='config.ini'):
        """Load data from source files into DB.

        Args:
            cfg_path (str): Path to the config file.
        """
        print('Reading config file:', os.path.abspath(cfg_path))
        cfg = ConfigParser(interpolation=ExtendedInterpolation())
        cfg.read(cfg_path)

        db_dir = cfg['.global']['db_dir']
        db_processor = DBProcessor(db_dir=db_dir)

        for section in cfg.sections():
            # ignore sections starting with a dot
            if not section.startswith('.'):
                # get corpus parser based on corpus name
                corpus_parser_class = CorpusParserMapper.map(section)
                data = dict(cfg.items(section))
                corpus_parser = corpus_parser_class(data)

                # get the corpus
                corpus = corpus_parser.parse()

                # add the corpus to the DB
                db_processor.insert_corpus(corpus)


def main():
    Loader.load()


if __name__ == "__main__":
    main()



import re

from acqdiv.parsers.toolbox.cleaners.cleaner import ToolboxCleaner
from acqdiv.parsers.corpora.main.tuatschin.reader \
    import TuatschinReader
from acqdiv.parsers.corpora.main.tuatschin.pos_mapper \
    import TuatschinPOSMapper
from acqdiv.parsers.corpora.main.tuatschin.gloss_mapper \
    import TuatschinGlossMapper


class TuatschinCleaner(ToolboxCleaner):

    @classmethod
    def remove_punctuation_utterance(cls, utterance):
        """Remove punctuation in the utterance.

        Utterances can contain the following punctuation: '?', '!', ',' and
        dot repetitions '. . .'.
        """
        utterance = re.sub(r'[.?!,]', '', utterance)
        return cls.remove_redundant_whitespaces(utterance)

    @staticmethod
    def unify_unknown_utterance(pos_tier):
        """Unify unknowns in the utterance.

        'xxx' stands for incomprehensible forms and is standardized to '???'.
        """
        return pos_tier.replace('xxx', '???')

    @classmethod
    def clean_utterance(cls, utterance):
        utterance = cls.remove_punctuation_utterance(utterance)
        utterance = cls.unify_unknown_utterance(utterance)
        return utterance

    # ---------- cross cleaners ----------

    @classmethod
    def unify_unknown_gloss_tier(cls, gloss_tier, seg_tier):
        """Unify unknowns on the gloss tier.

        Annotated with 'inv'. Since 'inv' (=invariable forms) is also used
        in other cases (such as ADP, ADV, etc.), the segment tier has to be
        used for inference.

        Returns:
            str: The cleaned gloss tier.
        """
        gloss_words = TuatschinReader.get_gloss_words(gloss_tier)
        seg_words = TuatschinReader.get_seg_words(seg_tier)

        if len(gloss_words) == len(seg_words):
            for i, seg_word in enumerate(seg_words):
                if seg_word == 'XXX':
                    gloss_words[i] = '???'

            return ' '.join(gloss_words)

        return gloss_tier

    @classmethod
    def remove_punct_inv(cls, gloss_tier, pos_tier):
        """Remove punctuation inv's on the gloss tier.

        Punctuation receives on the gloss tier the 'inv' value. Since
        'inv' (=invariable forms) is also used in other cases
        (such as ADP, ADV, etc.), the POS tier has to be used for inference.

        Returns:
            str: The cleaned gloss tier.
        """
        gloss_words = TuatschinReader.get_gloss_words(gloss_tier)
        pos_words = TuatschinReader.get_pos_words(pos_tier)

        if len(gloss_words) == len(pos_words):
            for i, pos_word in enumerate(pos_words):
                if pos_word == 'PUNCT':
                    gloss_words[i] = ''

            return ' '.join((w for w in gloss_words if w))

        return gloss_tier

    @classmethod
    def cross_clean(cls, rec_dict):
        gloss_tier = TuatschinReader.get_gloss_tier(rec_dict)
        seg_tier = TuatschinReader.get_seg_tier(rec_dict)
        pos_tier = TuatschinReader.get_pos_tier(rec_dict)
        gloss_tier = cls.remove_punct_inv(gloss_tier, pos_tier)
        gloss_tier = cls.unify_unknown_gloss_tier(gloss_tier, seg_tier)
        rec_dict['morphosyn'] = gloss_tier

        return rec_dict

    # ---------- seg tier cleaners ----------

    @staticmethod
    def unify_unknown_seg_tier(seg_tier):
        """Unify unknowns on the segment tier.

        'XXX' stands for unknown/incomprehensible forms and is standardized
        to '???'.
        """
        return seg_tier.replace('XXX', '???')

    @classmethod
    def remove_dot_repetitions_seg_tier(cls, seg_tier):
        """Remove trail offs on the segment tier.

        Trail offs are marked with dot repetitions (3 or 4 times).
        """
        seg_tier = re.sub(r'\. \. \.( \.)?', '', seg_tier)
        return cls.remove_redundant_whitespaces(seg_tier)

    @classmethod
    def remove_punctuation_seg_tier(cls, seg_tier):
        """Remove punctuation on the segment tier.

        The segment tier can contain the following punctuation:
        `?`, `!`, `,` and `.`
        """
        utterance = re.sub(r'[?!,.]', '', seg_tier)
        return cls.remove_redundant_whitespaces(utterance)

    @classmethod
    def clean_seg_tier(cls, seg_tier):
        for cleaning_method in [
            cls.remove_dot_repetitions_seg_tier,
            cls.remove_punctuation_seg_tier,
            cls.unify_unknown_seg_tier
        ]:
            seg_tier = cleaning_method(seg_tier)

        return seg_tier

    # ---------- POS tier cleaners ----------

    @classmethod
    def remove_punct_pos_tier(cls, pos_tier):
        """Remove PUNCT on the POS tier.

        Punctuation receives the POS tag 'PUNCT'.
        """
        pos_tier = pos_tier.replace('PUNCT', '')
        return cls.remove_redundant_whitespaces(pos_tier)

    @staticmethod
    def unify_unknown_pos_tier(pos_tier):
        """Unify unknowns on the POS tier.

        `X` stands either for child words that have an unclear POS tag or
        incomprehensible/unknown forms (see XXX on the segment tier) and is
        standardized to '???'. Sometimes, specifications (such as `_Chld`) are
        suffixed to `X`.
        """
        rgx = re.compile(r'\bX(_\S+)?\b')
        return rgx.sub('???', pos_tier)

    @classmethod
    def clean_pos_tier(cls, pos_tier):
        for cleaning_method in [
                cls.remove_punct_pos_tier,
                cls.unify_unknown_pos_tier
        ]:
            pos_tier = cleaning_method(pos_tier)

        return pos_tier

    # ---------- segment word cleaners ----------

    @classmethod
    def clean_seg_word(cls, segment_word):
        segment_word = cls.lowercase_seg_word(segment_word)
        return segment_word

    @classmethod
    def lowercase_seg_word(cls, segment_word):
        return segment_word.lower()

    # ---------- morpheme cleaners ----------

    @classmethod
    def clean_pos(cls, pos):
        return TuatschinPOSMapper.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return TuatschinPOSMapper.map(pos_ud, ud=True)

    @classmethod
    def clean_gloss(cls, gloss):
        return TuatschinGlossMapper.map(gloss)



from acqdiv.model.uniquespeaker import UniqueSpeaker

uniquespeakers = {}


def set_unique_speakers(corpus, speakers):
    """Set unique speakers.

    Args:
        corpus (str): The corpus name.
        speakers (List[acqdiv.model.speaker.Speaker]): The session speakers.
    """
    for speaker in speakers:
        key = (corpus, speaker.code, speaker.name, speaker.birth_date)

        if key in uniquespeakers:
            uspeaker = uniquespeakers[key]
        else:
            uspeaker = UniqueSpeaker()
            uspeaker.code = speaker.code
            uspeaker.name = speaker.name
            uspeaker.gender = speaker.gender
            uspeaker.gender_raw = speaker.gender_raw
            uspeaker.birth_date = speaker.birth_date
            uniquespeakers[key] = uspeaker

        speaker.uniquespeaker = uspeaker



import os

from acqdiv.parsers.chat.readers.reader import CHATReader
from acqdiv.parsers.chat.cleaners.cleaner import CHATCleaner
from acqdiv.parsers.session_parser import SessionParser

from acqdiv.util.age import get_age_in_days
from acqdiv.util.role import RoleMapper
from acqdiv.util.alignment import align_words_morphemes, fix_misalignments
from acqdiv.util.childdirectedness import infer_childdirected

from acqdiv.model.session import Session
from acqdiv.model.speaker import Speaker
from acqdiv.model.utterance import Utterance
from acqdiv.model.word import Word
from acqdiv.model.morpheme import Morpheme


class CHATParser(SessionParser):
    """Gathers all data for the DB for a given CHAT session file.

    Uses the CHATReader for reading and inferring data from the CHAT file and
    CHATCleaner for cleaning these data.
    """

    role_mapper = RoleMapper()

    def __init__(self, session_path):
        self.session_path = session_path
        self.session_filename = os.path.basename(self.session_path)

        with open(session_path) as session_file:
            self.reader = self.get_reader(session_file)

        self.cleaner = self.get_cleaner()
        self.consistent_actual_target = True

    @staticmethod
    def get_reader(session_file):
        """Return a CHATReader instance.

        Returns:
            acqdiv.parsers.chat.readers.reader.CHATReader:
            The reader instance.
        """
        return CHATReader(session_file)

    @staticmethod
    def get_cleaner():
        """Return a CHATCleaner instance.

        Returns:
            acqdiv.parsers.chat.cleaners.cleaner.CHATCleaner:
            The cleaner instance.
        """
        return CHATCleaner()

    def parse(self):
        """Get the session instance.

        Returns:
            acqdiv.model.session.Session: The Session instance.
        """
        session = Session()
        self.add_session_metadata(session)
        self.add_speakers(session)
        self.add_records(session)

        return session

    def add_session_metadata(self, session):
        """Add the metadata to the session."""
        date = self.cleaner.clean_date(self.reader.get_session_date())
        media_filename = self.reader.get_session_media_filename()

        # any corrections of the metadata
        date, media_filename = self.cleaner.clean_session_metadata(
            self.session_filename, date, media_filename)

        session.source_id = os.path.splitext(self.session_filename)[0]
        session.date = date if date else None
        session.media_filename = media_filename if media_filename else None

    def add_speakers(self, session):
        """Add the speakers to the session."""
        while self.reader.load_next_speaker():
            speaker = Speaker()

            speaker_label = self.reader.get_speaker_label()
            name = self.reader.get_speaker_name()
            role_raw = self.reader.get_speaker_role()
            age_raw = self.reader.get_speaker_age()
            gender_raw = self.reader.get_speaker_gender()
            language = self.reader.get_speaker_language()
            birth_date = self.cleaner.clean_date(
                                self.reader.get_speaker_birthdate())
            target_child = self.reader.get_target_child()

            # any corrections of the metadata
            speaker_label, name, role_raw, age_raw, gender_raw, language, birth_date = \
                self.cleaner.clean_speaker_metadata(
                    self.session_filename, speaker_label, name, role_raw, age_raw,
                    gender_raw, language, birth_date, target_child)

            speaker.code = speaker_label
            speaker.name = self.cleaner.clean_name(name)
            speaker.languages_spoken = language
            speaker.birth_date = birth_date

            speaker.age_raw = age_raw
            speaker.age = self.cleaner.clean_age(speaker.age_raw)
            speaker.age_in_days = get_age_in_days(speaker.age)

            speaker.role_raw = role_raw
            speaker.role = self.role_mapper.role_raw2role(speaker.role_raw)
            speaker.macro_role = self.role_mapper.infer_macro_role(
                speaker.role_raw, speaker.age_in_days, speaker.code)

            speaker.gender_raw = gender_raw
            speaker.gender = self.cleaner.clean_gender(speaker.gender_raw)
            if not speaker.gender:
                speaker.gender = self.role_mapper.role_raw2gender(
                    speaker.role_raw)

            session.speakers.append(speaker)

    def add_records(self, session):
        """Add the records."""
        while self.reader.load_next_record():
            utt = self.add_utterance(session)
            self.add_words(utt)
            self.add_morphemes(utt)
            align_words_morphemes(utt)

    @staticmethod
    def _get_speaker(label, speakers):
        for speaker in speakers:
            if speaker.code == label:
                return speaker

        return None

    def add_utterance(self, session):
        """Add the utterance to the session."""
        utt = Utterance()
        session.utterances.append(utt)
        utt.source_id = self.get_source_id()
        speaker_label = self.cleaner.clean_record_speaker_label(
            self.session_filename, self.reader.get_record_speaker_label())
        utt.speaker = self._get_speaker(speaker_label, session.speakers)
        addressee_label = self.cleaner.clean_record_speaker_label(
            self.session_filename, self.reader.get_addressee())
        utt.addressee = self._get_speaker(addressee_label, session.speakers)
        utt.childdirected = infer_childdirected(utt)
        utt.translation = self.cleaner.clean_translation(
            self.reader.get_translation())
        utt.comment = self.reader.get_comments()
        utt.utterance_raw = self.reader.get_utterance()
        utt.start_raw = self.reader.get_start_time()
        utt.start = self.cleaner.clean_timestamp(utt.start_raw)
        utt.end_raw = self.reader.get_end_time()
        utt.end = self.cleaner.clean_timestamp(utt.end_raw)
        utt.sentence_type = self.reader.get_sentence_type()
        utt.warning = ''

        # set raw morphology tiers
        utt.morpheme_raw = self.reader.get_seg_tier()
        utt.gloss_raw = self.reader.get_gloss_tier()
        utt.pos_raw = self.reader.get_pos_tier()

        # clean the morphology tiers
        seg_tier = self.cleaner.clean_seg_tier(utt.morpheme_raw)
        gloss_tier = self.cleaner.clean_gloss_tier(utt.gloss_raw)
        pos_tier = self.cleaner.clean_pos_tier(utt.pos_raw)

        # actual & target distinction
        actual_utt = self.cleaner.clean_utterance(
            self.reader.get_actual_utterance())
        target_utt = self.cleaner.clean_utterance(
            self.reader.get_target_utterance())

        # cross cleaning
        actual_utt, target_utt, seg_tier, gloss_tier, pos_tier = \
            self.cleaner.utterance_cross_clean(
                utt.utterance_raw, actual_utt, target_utt,
                seg_tier, gloss_tier, pos_tier)

        utt.actual_utterance = actual_utt
        utt.target_utterance = target_utt
        utt.morpheme = seg_tier
        utt.gloss = gloss_tier
        utt.pos = pos_tier

        return utt

    def add_words(self, utt):
        """Add the words to the utterance."""
        actual_words = self.reader.get_utterance_words(utt.actual_utterance)
        target_words = self.reader.get_utterance_words(utt.target_utterance)

        for word_actual, word_target in zip(actual_words, target_words):

            w = Word()
            utt.words.append(w)

            if self.reader.get_standard_form() == 'actual':
                word = word_actual
            else:
                word = word_target

            w.word_language = self.reader.get_word_language(word)
            w.word = self.cleaner.clean_word(word)
            w.word_actual = self.cleaner.clean_word(word_actual)
            w.word_target = self.cleaner.clean_word(word_target)
            w.warning = ''

            if not self.consistent_actual_target:
                if word_actual == word_target:
                    w.word_actual = None
                    w.word_target = None

        # rebuild utterance from cleaned words
        utt.utterance = ' '.join(w.word for w in utt.words)

    def get_source_id(self):
        """Get the source id of the current utterance."""
        fname = self.session_path.split('/')[-1]
        fname_no_ext = fname.split('.')[0]
        uid = self.reader.get_uid()
        if fname:
            return '{}_{}'.format(fname_no_ext, uid)
        return uid

    def add_morphemes(self, utt):
        """Add the morphemes to the utterance."""
        wsegs = self.reader.get_seg_words(utt.morpheme)
        wglosses = self.reader.get_gloss_words(utt.gloss)
        wposes = self.reader.get_pos_words(utt.pos)

        if self.reader.get_main_morpheme() == 'segment':
            wsegs, wglosses, wposes = \
                fix_misalignments([wsegs, wglosses, wposes])
        else:
            wglosses, wsegs, wposes = \
                fix_misalignments([wglosses, wsegs, wposes])

        # go through all morpheme words
        for wseg, wgloss, wpos in zip(wsegs, wglosses, wposes):

            cleaned_wseg = self.cleaner.clean_seg_word(wseg)
            cleaned_wgloss = self.cleaner.clean_gloss_word(wgloss)
            cleaned_wpos = self.cleaner.clean_pos_word(wpos)

            # collect morpheme data of a word
            wmorphemes = []

            # get morphemes from the morpheme words
            segments = self.reader.get_segments(cleaned_wseg)
            glosses = self.reader.get_glosses(cleaned_wgloss)
            poses = self.reader.get_poses(cleaned_wpos)

            # determine number of morphemes to be considered
            if self.reader.get_main_morpheme() == 'segment':
                segments, glosses, poses = \
                    fix_misalignments([segments, glosses, poses])
            else:
                glosses, segments, poses = \
                    fix_misalignments([glosses, segments, poses])

            # go through morphemes
            for seg, gloss, pos in zip(segments, glosses, poses):
                m = Morpheme()

                m.morpheme_language = self.reader.get_morpheme_language(
                    seg, gloss, pos)

                m.morpheme = self.cleaner.clean_segment(seg)
                m.gloss_raw = self.cleaner.clean_gloss_raw(gloss)
                m.gloss = self.cleaner.clean_gloss(gloss)
                m.pos_raw = self.cleaner.clean_pos_raw(pos)
                m.pos = self.cleaner.clean_pos(pos)
                m.pos_ud = self.cleaner.clean_pos_ud(pos)
                m.type = self.reader.get_morpheme_type()

                wmorphemes.append(m)

            utt.morphemes.append(wmorphemes)



import re

from acqdiv.parsers.chat.cleaners.cleaner import CHATCleaner
from acqdiv.parsers.chat.cleaners.utterance_cleaner \
    import CHATUtteranceCleaner
from acqdiv.parsers.corpora.main.sesotho.gloss_mapper \
    import SesothoGlossMapper
from acqdiv.parsers.corpora.main.sesotho.pos_mapper \
    import SesothoPOSMapper


class SesothoCleaner(CHATCleaner):

    @staticmethod
    def correct_birthdate(date):
        if date == '1984-01-01':
            return ''

        return date

    @staticmethod
    def correct_speaker_name(name, speaker_label):
        if name == 'Khetheng':
            return 'Khethang'

        if name == 'child':
            return 'Unidentified_child'

        if name == 'MantSo':
            return 'Mantso'

        if name == 'NtsoakiCousin':
            return 'Ntsoaki_Cousin'

        if name == 'TlalaneCousin':
            return 'Tlalane_Cousin'

        if speaker_label == 'MAN':
            return 'Mathoto'

        if speaker_label == 'JUL':
            return 'Julia'

        if speaker_label == 'LIN':
            return 'Lineo'

        if speaker_label == 'NEU':
            return 'Neuoe_Cousin'

        return name

    @classmethod
    def clean_speaker_metadata(
            cls, session_filename, label, name, role,
            age, gender, language, birth_date, target_child):
        """Correct label, role and name of speaker."""
        birth_date = cls.correct_birthdate(birth_date)
        name = cls.correct_speaker_name(name, label)

        return label, name, role, age, gender, language, birth_date

    # ---------- utterance cleaning ----------

    @classmethod
    def clean_utterance(cls, utterance):
        utterance = cls.remove_words_in_parentheses(utterance)
        utterance = cls.remove_parentheses(utterance)
        return super().clean_utterance(utterance)

    @staticmethod
    def remove_words_in_parentheses(utterance):
        """Remove words in parentheses.

        In Sesotho, these are only used to mark contractions of the
        verb 'go' which are conventional in both child and adult
        speech.
        """
        if utterance.startswith('('):
            return re.sub(r'\(\w+\) ', '', utterance)

        return re.sub(r' \(\w+\) ', ' ', utterance)

    @staticmethod
    def remove_parentheses(utterance):
        """Remove parentheses.

        Because words that are entirely surrounded by parentheses are
        already removed, this method should just remove parentheses,
        that only surround a part of a word.

        Such parentheses are leftovers from the morpheme joining.
        """
        return re.sub(r'[()]', '', utterance)

    @classmethod
    def clean_translation(cls, translation):
        """Clean the Sesotho translation tier."""
        return cls.remove_timestamp(translation)

    @classmethod
    def remove_timestamp(cls, translation):
        """Remove timestamps in the Sesotho translation tier."""
        translation = re.sub(r'[0-9]+_[0-9]+', '', translation)
        return CHATUtteranceCleaner.remove_redundant_whitespaces(translation)

    # ---------- cross cleaning ----------

    @classmethod
    def utterance_cross_clean(
            cls, raw_utt, actual_utt, target_utt,
            seg_tier, gloss_tier, pos_tier):
        """Clean seg_tier, gloss_tier and pos_tier from contractions."""
        seg_tier, gloss_tier, pos_tier = cls.remove_contractions(
            seg_tier, gloss_tier, pos_tier)
        return actual_utt, target_utt, seg_tier, gloss_tier, pos_tier

    @classmethod
    def remove_contractions(cls, seg_tier, gloss_tier, pos_tier):
        """Remove contractions

        Remove words on the morpheme tiers, that are fully surrounded
        by parentheses. These parentheses can be used in Sesotho to mark
        contractions of the verb go.

        Since such contractions are only marked on the segment tier and
        misalignments should be avoided, the pos-words and gloss-words
        at the same index are also deleted.
        """
        gloss_tier = re.sub(r'\s+,\s+', ',', gloss_tier)
        pos_tier = re.sub(r'\s+,\s+', ',', pos_tier)
        seg_words = seg_tier.split(' ')
        gloss_words = gloss_tier.split(' ')
        pos_words = pos_tier.split(' ')
        seg_words_clean = []
        gloss_words_clean = []
        pos_words_clean = []
        slen = len(seg_words)
        glen = len(gloss_words)
        plen = len(pos_words)

        for i in range(glen):
            # Check if i is in range of seg_words to then check if there
            # is a contraction.
            if i < slen:
                if not re.search(r'^\(.*\)$', seg_words[i]):
                    # i must be in range for gloss_words and seg_words,
                    # but check if i is in range for pos_words.
                    gloss_words_clean.append(gloss_words[i])
                    seg_words_clean.append(seg_words[i])
                    if i < plen:
                        pos_words_clean.append(pos_words[i])
            else:
                # If i not in range for seg_words, check and append for
                # the other tiers, since there could be a misalignment.
                gloss_words_clean.append(gloss_words[i])
                if i < plen:
                    pos_words_clean.append(pos_words[i])

        seg_tier = ' '.join(seg_words_clean)
        gloss_tier = ' '.join(gloss_words_clean)
        pos_tier = ' '.join(pos_words_clean)

        return seg_tier, gloss_tier, pos_tier

    # ---------- morpheme cleaning ----------

    @classmethod
    def clean_seg_tier(cls, seg_tier):
        """Clean the segment tier by removing the terminator."""
        return CHATUtteranceCleaner.remove_terminator(seg_tier)

    @classmethod
    def clean_gloss_tier(cls, gloss_tier):
        """Clean the gloss tier."""
        for method in [CHATUtteranceCleaner.remove_terminator,
                       cls.remove_spaces_noun_class_parentheses,
                       cls.replace_noun_class_separator]:
            gloss_tier = method(gloss_tier)

        return gloss_tier

    @staticmethod
    def remove_spaces_noun_class_parentheses(gloss_tier):
        """Remove spaces in noun class parentheses.

        Noun classes in Sesotho are indicated as '(x , y)'. The spaces
        around the comma are removed so that word splitting by space
        doesn't split noun classes.
        """
        return re.sub(r'\s+,\s+', ',', gloss_tier)

    @staticmethod
    def replace_noun_class_separator(gloss_tier):
        """Replace '/' as noun class separator with '|'.

        This is to ensure that '/' can't be confused with '/' as a
        morpheme separator.
        """
        return re.sub(r'(\d+a?)/(\d+a?)', r'\1|\2', gloss_tier)

    @classmethod
    def clean_pos_tier(cls, pos_tier):
        """Clean pos_tier with same methods as gloss_tier."""
        return cls.clean_gloss_tier(pos_tier)

    @classmethod
    def clean_seg_word(cls, seg_word):
        """Remove parentheses."""
        return re.sub(r'[()]', '', seg_word)

    @classmethod
    def replace_concatenators(cls, gloss_word):
        """Replace '_' as concatenator of glosses with '.'.

        But don't replace '_' as concatenator of
        verb-multi-word-expressions.
        """
        glosses_raw = gloss_word.split('-')
        glosses_clean = []
        pos = ''
        passed_stem = False
        len_raw = len(glosses_raw)
        for gloss in glosses_raw:
            if len_raw == 1 or (re.search(r'(v|id)\^|\(\d', gloss)
                                       or re.match(r'(aj$|nm$|ps\d+)', gloss)):
                passed_stem = True
            elif not passed_stem:
                pos = 'pfx'
            elif passed_stem:
                pos = 'sfx'
            if pos == 'sfx' or pos == 'pfx':
                if not re.search(r'[vs]\^', gloss):
                    if not re.search(r'\(\d+', gloss):
                        glosses_clean.append(re.sub(r'_', r'.', gloss))
                    else:
                        glosses_clean.append(gloss)
                else:
                    glosses_clean.append(gloss)
            else:
                glosses_clean.append(gloss)

        gloss_word = '-'.join(glosses_clean)
        return gloss_word

    @staticmethod
    def remove_parentheses_inf(gloss_word):
        """Remove parentheses from infinitives.

        In Sesotho some infinitives are partially surrounded by
        parentheses. Remove those parentheses.
        """
        if not re.search(r'^\(.*\)$', gloss_word):
            return re.sub(r'\(([a-zA-Z]\S+)\)', r'\1', gloss_word)

        return gloss_word

    @classmethod
    def clean_gloss_word(cls, gloss_word):
        """Clean a Sesotho gloss word."""
        gloss_word = cls.replace_concatenators(gloss_word)
        gloss_word = cls.remove_parentheses_inf(gloss_word)
        return super().clean_gloss_word(gloss_word)

    @classmethod
    def clean_gloss(cls, gloss):
        return SesothoGlossMapper.map(gloss)

    @classmethod
    def clean_pos(cls, pos):
        return SesothoPOSMapper.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return SesothoPOSMapper.map(pos_ud, ud=True)



from acqdiv.parsers.toolbox.readers.reader import ToolboxReader


class QaqetReader(ToolboxReader):

    @classmethod
    def get_addressee(cls, rec):
        add = rec.get('addr', '')

        if add == 'unknown':
            return ''

        return add

    @classmethod
    def get_actual_utterance(cls, rec):
        return rec.get('tx', '')

    @classmethod
    def get_target_utterance(cls, rec_dict):
        return rec_dict.get('trs-i', '')

    @classmethod
    def get_translation(cls, rec):
        return rec.get('ft', '')

    @classmethod
    def get_comment(cls, rec):
        return rec.get('nt', '')

    @classmethod
    def get_sentence_type(cls, rec):
        """Get sentence type.

        Usually, there are no terminators in the transcriptions. In this case,
        the default value used. Only questions are explicitly marked,
        exclamations not.
        """
        utterance = cls.get_actual_utterance(rec)
        if utterance.endswith('?'):
            return 'question'
        else:
            return 'default'

    @staticmethod
    def get_morpheme_type():
        return 'actual'



from acqdiv.parsers.corpus_parser import CorpusParser
from acqdiv.parsers.corpora.main.japanese_miipro.session_parser \
    import JapaneseMiiProSessionParser


class JapaneseMiiProCorpusParser(CorpusParser):

    def get_session_parser(self, session_path):
        return JapaneseMiiProSessionParser(session_path)



import re

from acqdiv.parsers.toolbox.cleaners.cleaner import ToolboxCleaner
from acqdiv.parsers.corpora.main.indonesian.gloss_mapper \
    import IndonesianGlossMapper
from acqdiv.parsers.corpora.main.indonesian.pos_mapper \
    import IndonesianPOSMapper


class IndonesianCleaner(ToolboxCleaner):

    @classmethod
    def remove_utterance_punctuation(cls, utterance):
        utterance = re.sub('[‘’\'“”\".!,;:+/]|\?$|<|>', '', utterance)
        return cls.remove_redundant_whitespaces(utterance)

    @classmethod
    def remove_insecure_transcription_marker(cls, utterance):
        return re.sub('\[\?\]', '', utterance)

    @classmethod
    def clean_utterance(cls, utterance):
        for cleaning_method in [
                super().clean_utterance, cls.remove_utterance_punctuation,
                cls.remove_insecure_transcription_marker]:
            utterance = cleaning_method(utterance)

        return utterance

    @staticmethod
    def remove_morph_tier_punctuation(morpheme_tier):
        return re.sub('[‘’\'“”\".!,:?+/]', '', morpheme_tier)

    @staticmethod
    def unify_unknown(utterance):
        return re.sub('xxx?|www|0', '???', utterance)

    @classmethod
    def clean_morph_tier(cls, morph_tier):
        for cleaning_method in [
                cls.remove_morph_tier_punctuation, cls.unify_unknown]:
            morph_tier = cleaning_method(morph_tier)
        return morph_tier

    @classmethod
    def clean_gloss(cls, gloss):
        return IndonesianGlossMapper.map(gloss)

    @classmethod
    def clean_pos(cls, pos):
        return IndonesianPOSMapper.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return IndonesianPOSMapper.map(pos_ud, ud=True)



import os
import re

from lxml import objectify


class MetadataParser(object):
    """ Base metadata parser class.

    metadata.MetadataParser wraps an lxml.objectify tree. It gets relevant metadata
    from the tree and exposes them in a dictionary.

    Attributes:
        metadata: A dictionary of all metadata values extracted from the
                  file the tree was initialized with.
    """

    def __init__(self, path):
        self.path = path
        self.tree = objectify.parse(path)
        self.root = self.tree.getroot()
        self.metadata = {
            '__attrs__': self.parse_attrs(self.root),
                    }
        self.metadata['__attrs__']['Cname'] = re.sub(r'\.xml.*|\.imdi.*', "", os.path.basename(str(self.path)))

        # Special case for Indonesian
        # Explanation: this converts the session ID to the same format as in the body files
        # Unless that issue was fixed in another way we will probably still want it

        # TODO: figure out what's going wrong with Indonesian
        """
        print(self.metadata['__attrs__']['Cname'])

        if self.metadata_path['corpus'] == "Indonesian":
            match = re.match(r"(\w{3})-\d{2}(\d{2})-(\d{2})-(\d{2})",self.metadata['__attrs__']['Cname'])
            self.metadata['__attrs__']['Cname'] = match.group(1).upper() + '-' + match.group(4) + match.group(3) + match.group(2)
        self.metadata['__attrs__']['schemaLocation'] = self.metadata['__attrs__'].pop('{http://www.w3.org/2001/XMLSchema-instance}schemaLocation')
        """
        # assert existing_dir(self.input_path())


    def parse_attrs(self, e):
        """
        Parses and returns attributes of the root node.

        args:
            e: root node of an XML session file

        returns:
            A dictionary of all attributes of e.
        """
        return {k: e.attrib[k] for k in e.keys()}


    def get_everything(self, root):
        """
        Parses and returns a list of all tag tuples in the tree.

        args:
            root: Root node of an XML session file

        returns:
            a list of tuples, each representing all tags of a node
        """
        everything = []
        for e in root.getiterator():
            if e is not None:
                everything.append((e.tag.replace("{http://www.mpi.nl/IMDI/Schema/IMDI}", ""), e))
        return everything


from pathlib import Path

from acqdiv.parsers.corpus_parser import CorpusParser
from acqdiv.parsers.corpora.main.russian.session_parser \
    import RussianSessionParser


class RussianCorpusParser(CorpusParser):

    def get_session_parser(self, session_path):
        metadata_filename = Path(session_path).with_suffix('.imdi').name
        metadata_filepath = Path(self.cfg['metadata_dir']) / metadata_filename

        return RussianSessionParser(session_path, str(metadata_filepath))



from acqdiv.parsers.corpus_parser import CorpusParser
from acqdiv.parsers.corpora.main.japanese_miyata.session_parser \
    import JapaneseMiyataSessionParser


class JapaneseMiyataCorpusParser(CorpusParser):

    def get_session_parser(self, session_path):
        return JapaneseMiyataSessionParser(session_path)



import re

from acqdiv.parsers.toolbox.cleaners.cleaner import ToolboxCleaner
from acqdiv.parsers.corpora.main.russian.gloss_mapper \
    import RussianGlossMapper
from acqdiv.parsers.corpora.main.russian.pos_mapper \
    import RussianPOSMapper


class RussianCleaner(ToolboxCleaner):

    @classmethod
    def remove_punctuation(cls, utterance):
        utterance = re.sub(
            '[‘’\'“”\".!,:+/]+|(&lt; )|(?<=\\s)\?(?=\\s|$)', '', utterance)
        return cls.remove_redundant_whitespaces(utterance)

    @classmethod
    def remove_dashes(cls, utterance):
        return re.sub('\\s-\\s', ' ', utterance)

    @classmethod
    def remove_insecure_transcription_markers(cls, utterance):
        """Remove insecure transcription markers.

        Insecure transcription markers: [?], [=( )?], [xxx].
        Note that [xxx] usually replaces a complete utterance and is
        non-aligned, in contrast to xxx without brackets, which can be
        counted as a word.
        """
        # TODO: Get warnings on utterance level
        if re.search('\[(\s*=?.*?|\s*xxx\s*)\]', utterance):
            utterance = re.sub('\[\s*=?.*?\]', '', utterance)
            return cls.remove_redundant_whitespaces(utterance)

        return utterance

    @classmethod
    def remove_equal_signs(cls, utterance):
        utterance = utterance.replace('=', '')
        return cls.remove_redundant_whitespaces(utterance)

    @classmethod
    def clean_utterance(cls, utterance):
        for cleaning_method in [
                super().clean_utterance, cls.remove_punctuation,
                cls.remove_dashes, cls.remove_insecure_transcription_markers,
                cls.remove_equal_signs]:
            utterance = cleaning_method(utterance)

        return utterance

    @staticmethod
    def remove_seg_punctuation(seg_tier):
        return re.sub('[‘’\'“”\".!,:\-?+/]', '', seg_tier)

    @staticmethod
    def unify_unknown(seg_tier):
        return re.sub('xxx?|www', '???', seg_tier)

    @classmethod
    def clean_seg_tier(cls, seg_tier):
        for cleaning_method in [cls.remove_seg_punctuation, cls.unify_unknown]:
            seg_tier = cleaning_method(seg_tier)

        return seg_tier

    @staticmethod
    def clean_gloss_pos_punctuation(gloss_pos_tier):
        return gloss_pos_tier.replace('PUNCT', '').replace('ANNOT', '').\
            replace('<NA: lt;> ', '')

    @classmethod
    def clean_gloss_tier(cls, gloss_tier):
        return cls.clean_gloss_pos_punctuation(gloss_tier)

    @classmethod
    def clean_pos_tier(cls, pos_tier):
        return cls.clean_gloss_pos_punctuation(pos_tier)

    @classmethod
    def clean_lang_tier(cls, lang_tier):
        return cls.clean_gloss_pos_punctuation(lang_tier)

    @classmethod
    def clean_gloss(cls, gloss):
        return RussianGlossMapper.map(gloss)

    @classmethod
    def clean_pos(cls, pos):
        return RussianPOSMapper.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return RussianPOSMapper.map(pos_ud, ud=True)



import csv

from acqdiv.util.path import get_full_path


def _parse():
    """Return durations from session_durations.csv.

    Returns:
        Dict[Dict[str]]: The duration indexed by corpus and source_id.
    """
    durations_csv_path = 'util/resources/session_durations.csv'
    full_path = get_full_path(durations_csv_path)

    durations = {}

    with open(full_path, 'r', encoding='utf8') as f:
        reader = csv.DictReader(f)

        for row in reader:
            corpus = row['corpus']
            source_id = row['source_id']
            duration = row['duration']

            if corpus in durations:
                durations[corpus][source_id] = duration
            else:
                durations[corpus] = {source_id: duration}

    return durations


_durations = _parse()


def extract_duration(corpus, source_id):
    """Get the duration.

    Args:
        corpus (str): The corpus name.
        source_id (str): The session name.

    Returns:
        str: The duration.
    """
    if corpus in _durations and source_id in _durations[corpus]:
        return _durations[corpus][source_id]

    return ''





class CHAT:

    def __init__(self):
        """
        date (str): The value at @Date.
        participants (Dict[str, acqdiv.parsers.chat.model.Participant]):
            The participants as read from @Participants, @ID, @Birth of.
        pid (str): The value at @PID.
        media_filename (str): The filename at @Media.
        media_format (str): The format at @Media.
        media_comment (str): The comment at @Media.
        records (List[acqdiv.parsers.chat.model.Record]): The records.
        """
        self.date = ''
        self.participants = {}
        self.pid = ''
        self.media_filename = ''
        self.media_format = ''
        self.media_comment = ''
        self.records = []



import re

from acqdiv.parsers.chat.cleaners.cleaner import CHATCleaner
from acqdiv.parsers.chat.cleaners.utterance_cleaner \
    import CHATUtteranceCleaner
from acqdiv.parsers.corpora.main.japanese_miipro.gloss_mapper \
    import JapaneseMiiProGlossMapper as GMp
from acqdiv.parsers.corpora.main.japanese_miipro.pos_mapper \
    import JapaneseMiiProPOSMapper as PMp


class JapaneseMiiProCleaner(CHATCleaner):

    @staticmethod
    def correct_speaker_label(session_filename, speaker_label):
        """Replace `CHI` label of target child."""
        if speaker_label == 'CHI':
            if session_filename.startswith('Asato'):
                return 'ALS'
            elif session_filename.startswith('Arika'):
                return 'APR'
            elif session_filename.startswith('Nanami'):
                return 'NJD'
            elif session_filename.startswith('Tomito'):
                return 'TOM'

        return speaker_label

    @classmethod
    def clean_record_speaker_label(cls, session_filename, speaker_label):
        speaker_label = cls.correct_speaker_label(
            session_filename, speaker_label)
        return speaker_label

    # ---------- morphology tier cleaning ----------

    @classmethod
    def remove_non_words(cls, morph_tier):
        """Remove all non-words from the morphology tier.

        Non-words have the POS tag 'tag'.
        """
        non_words_regex = re.compile(r'tag\|\S+')
        morph_tier = non_words_regex.sub('', morph_tier)
        return CHATUtteranceCleaner.remove_redundant_whitespaces(morph_tier)

    @classmethod
    def clean_morph_tier(cls, morph_tier):
        morph_tier = CHATUtteranceCleaner.remove_terminator(morph_tier)
        return cls.remove_non_words(morph_tier)

    # ---------- speaker metadata cleaning ----------

    @classmethod
    def clean_speaker_metadata(
            cls, session_filename, label, name, role,
            age, gender, language, birth_date, target_child):
        """Correct label, role and name of speaker."""

        tc_label, _ = target_child
        tc_label = cls.correct_speaker_label(session_filename, tc_label)
        label = cls.correct_speaker_label(session_filename, label)
        role = cls.correct_role(label, role, tc_label)
        name = cls.correct_name(session_filename, label, name)

        return label, name, role, age, gender, language, birth_date

    @staticmethod
    def correct_role(label, role, tc_label):
        """Correct role of target child.

        Returns:
            str: The corrected role.
        """
        if role == 'Child' and label == tc_label:
            return 'Target_Child'

        return role

    @staticmethod
    def correct_name(session_filename, label, name):
        """Correct name of speaker.

        Returns:
            str: The corrected name.
        """
        if session_filename.startswith('Asato'):
            if name == 'Asatokun':
                return 'Asato'
            elif label == 'APR':
                return 'Arika'
            elif label == 'MOT':
                return 'Mother_of_Arika_and_Asato'
            elif label == 'ALS':
                return 'Asato'

        elif session_filename.startswith('Arika'):
            if name == 'Arichan':
                return 'Arika'
            elif label == 'MOT':
                return 'Mother_of_Arika_and_Asato'
            elif name == 'Asatokun':
                return 'Asato'
            elif label == 'APR':
                return 'Arika'
            elif label == 'BAA':
                return 'Obaachan'

        elif session_filename.startswith('Nanami'):
            if name == 'Natchan':
                return 'Nanami'
            elif name == 'Fuyumichan':
                return 'Fuyumi'
            elif label == 'MOT':
                return 'Mother_of_Nanami'
            elif name == 'Arikachan':
                return 'Arika'
            elif name == 'Arichan':
                return 'Arika'
            elif name == 'Kantakun':
                return 'Kanta'
            elif name == 'Totchan':
                return 'Tomito'
            elif label == 'MTO':
                return 'Mother_of_Tomito'
            elif label == 'NJD':
                return 'Nanami'
            elif label == 'TMO' and name == "Totchan's_Mother":
                name = 'Mother_of_Tomito'

        elif session_filename.startswith('Tomito'):
            if name == 'Honokachan':
                return 'Honoka'
            elif name == 'Tomitokun':
                return 'Tomito'
            elif name == 'Totchan':
                return 'Tomito'
            elif label == 'APR':
                return 'Arika'
            elif label == 'MOT':
                return 'Mother_of_Tomito'

        return name

    # ---------- utterance cross clean ----------

    @classmethod
    def utterance_cross_clean(
            cls, raw_utt, actual_utt, target_utt,
            seg_tier, gloss_tier, pos_tier):

        seg_tier = cls.add_repetitions(raw_utt, seg_tier)
        gloss_tier = cls.add_repetitions(raw_utt, gloss_tier)
        pos_tier = cls.add_repetitions(raw_utt, pos_tier)

        seg_tier = cls.add_retracings(raw_utt, actual_utt, seg_tier)
        gloss_tier = cls.add_retracings(raw_utt, actual_utt, gloss_tier)
        pos_tier = cls.add_retracings(raw_utt, actual_utt, pos_tier)

        return actual_utt, target_utt, seg_tier, gloss_tier, pos_tier

    @classmethod
    def add_repetitions(cls, raw_utt, morph_tier):
        """Add repetitions to morphology tier."""
        # check if there are any repetitions
        if '[x ' in raw_utt:

            # execute same cleaning steps except those for scoped symbols
            for cleaning_method in [
                    CHATUtteranceCleaner.remove_terminator,
                    CHATUtteranceCleaner.unify_untranscribed,
                    CHATUtteranceCleaner.remove_events,
                    CHATUtteranceCleaner.remove_omissions,
                    CHATUtteranceCleaner.remove_linkers,
                    CHATUtteranceCleaner.remove_separators,
                    CHATUtteranceCleaner.remove_ca,
                    CHATUtteranceCleaner.remove_pauses_between_words,
                    CHATUtteranceCleaner.remove_commas,
                    CHATUtteranceCleaner.null_event_utterances
                    ]:
                raw_utt = cleaning_method(raw_utt)

            # remove scoped symbols except for repetitions
            scope_regex = re.compile(r'\[[^x].*?\]')
            raw_utt = scope_regex.sub('', raw_utt)
            raw_utt = CHATUtteranceCleaner.remove_redundant_whitespaces(
                raw_utt)

            # get words from utterance and morphology tier
            utt_words = re.split(r'(?<!\[x) (?!\[x)', raw_utt)
            morph_words = morph_tier.split(' ')

            # check for misalignments
            if len(utt_words) == len(morph_words):

                morph_new = []
                group = []
                for uw, mw in zip(utt_words, morph_words):

                    morph_new.append(mw)
                    match = re.search(r'\[x (\d+)', uw)

                    if uw.startswith('<'):
                        group = [mw]
                    elif match:
                        reps = int(match.group(1))
                        if group:
                            group.append(mw)
                            morph_new += (reps-1)*group
                            group = []
                        else:
                            morph_new += (reps-1)*[mw]
                    elif group:
                        group.append(mw)

                return ' '.join(morph_new)

        return morph_tier

    @classmethod
    def add_retracings(cls, raw_utt, actual_utt, morph_tier):
        """Add retracings to morphology tiers.

        Uses a fuzzy matching approach as the retracings are not always used
        correctly in the corpus. For example, in some cases they are used for
        repetitions: < soo soo > [/] soo.

        Only considers retracings of up to 2 words.
        """
        # only perform steps if there are retracings
        if '[/]' in raw_utt and morph_tier:

            regex = re.compile(r'((\S+)( \2)+)|((\S+) (\S+)( \5 \6)+)')
            actual_utt = ' '.join(
                [cls.clean_word(word) for word in actual_utt.split(' ')])
            repeated_words = list(regex.finditer(actual_utt))

            morph_words = morph_tier.split(' ')
            new = []
            for i, mword in enumerate(morph_words):

                new.append(mword)

                if repeated_words:
                    repeated = repeated_words[0]

                    if repeated.group(1):
                        wword = repeated.group(2)
                        # first three letters have to match
                        if wword[:3] in mword:
                            n_reps = len(repeated.group(1).split(' ')) - 1
                            new += n_reps*[mword]
                            del repeated_words[0]

                    elif i > 0 and repeated.group(4):
                        wword1 = repeated.group(5)
                        wword2 = repeated.group(6)

                        if (wword1[:3] in morph_words[i-1]
                                and wword2[:3] in mword):
                            n_reps = len(
                                repeated.group(7).lstrip(' ').split(' ')) - 1
                            new += n_reps*(morph_words[i-1], mword)

            return ' '.join(new)

        return morph_tier

    # ---------- morpheme cleaning ----------

    @classmethod
    def clean_gloss(cls, gloss):
        return GMp.map(gloss)

    @classmethod
    def clean_pos(cls, pos):
        return PMp.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return PMp.map(pos_ud, ud=True)



import re
import mmap
import contextlib

from acqdiv.parsers.toolbox.cleaners.cleaner import ToolboxCleaner
from acqdiv.parsers.toolbox.model.toolbox import ToolboxFile
from acqdiv.parsers.toolbox.model.record import Record

@contextlib.contextmanager
def memorymapped(path, access=mmap.ACCESS_READ):
    """ Return a block context with path as memory-mapped file. """
    fd = open(path)
    try:
        m = mmap.mmap(fd.fileno(), 0, access=access)
    except Exception:
        fd.close()
        raise
    try:
        yield m
    finally:
        m.close()
        fd.close()


class ToolboxFileParser:
    """Methods for creating a ToolboxFile instance."""

    @classmethod
    def parse(cls, path, separator):
        """Get a ToolboxFile instance.

        Args:
            path (str): Path to Toolbox file.
            separator (str): Field that marks beginning of a record.

        Returns:
            acqdiv.parsers.toolbox.model.toolbox.ToolboxFile: The Toolbox
            instance.
        """
        toolbox = ToolboxFile()

        with open(path, 'rb') as f:
            for record in cls.iter_records(f, separator):
                rec_dict = cls.get_record_dict(record)
                rec = Record()
                rec.tiers = rec_dict
                toolbox.records.append(rec)

        return toolbox

    @classmethod
    def iter_records(cls, toolbox_file, separator):
        """Iter the records of a toolbox file.

        Args:
            toolbox_file (file/file-like): The toolbox file.
            separator (str): Fieldmarker that marks beginning of record.

        Yields:
            str: The record.
        """
        _record_marker = re.compile(separator)

        with contextlib.closing(mmap.mmap(toolbox_file.fileno(),
                                          0, access=mmap.ACCESS_READ)) as data:
            ma = _record_marker.search(data)
            pos = ma.start()
            for ma in _record_marker.finditer(data, ma.end()):
                yield data[pos:ma.start()].decode()
                pos = ma.start()

            if ma is None:
                raise StopIteration
            else:
                yield data[pos:].decode()

    @classmethod
    def get_record_dict(cls, record):
        """Get the record dictionary.

        Metadata is ignored and returned as an empty dictionary.

        Args:
            record (str): Toolbox record.

        Returns:
            dict: Key and content of tiers.
        """
        rec_dict = {}

        # iter tiers of the record
        for tier in cls.get_tiers(record):
            # get field marker and content of tier
            field_marker, content = cls.get_tier(tier)
            # clean the content
            content = ToolboxCleaner.remove_redundant_whitespaces(content)

            # add content to dictionary
            rec_dict[field_marker] = content

        return rec_dict

    @staticmethod
    def get_tiers(record):
        """Return tiers of the record.

        Args:
            record (str): The record.

        Returns:
            list: The tiers of the record.
        """
        return record.split('\n')

    @staticmethod
    def get_tier(tier):
        """Return field marker and content.

        Args:
            tier (str): '\\name content'.

        Returns:
            tuple: (name, content).
        """
        # split into field marker and content
        tokens = re.split(r'\s+', tier, maxsplit=1)
        # get field marker
        field_marker = tokens[0]
        # remove \\ before the field marker
        field_marker = field_marker.replace("\\", "")

        # if content is missing
        if len(tokens) <= 1:
            content = ''
        else:
            content = tokens[1]

        return field_marker, content



from acqdiv.parsers.corpora.main.qaqet.imdi_parser import QaqetIMDI
from acqdiv.parsers.corpora.main.qaqet.reader import QaqetReader
from acqdiv.parsers.corpora.main.qaqet.cleaner import QaqetCleaner
from acqdiv.parsers.toolbox.parser import ToolboxParser
from acqdiv.model.word import Word


class QaqetSessionParser(ToolboxParser):

    def get_record_reader(self):
        return QaqetReader()

    def get_metadata_reader(self):
        return QaqetIMDI(self.metadata_path)

    def get_cleaner(self):
        return QaqetCleaner()

    def add_words(self, actual_utterance, target_utterance):
        utt = self.session.utterances[-1]

        actual_words = self.record_reader.get_words(actual_utterance)
        target_words = self.record_reader.get_words(target_utterance)

        if len(actual_words) != len(target_words):
            target_words = len(actual_words)*['']

        for actual, target in zip(actual_words, target_words):
            w = Word()
            utt.words.append(w)

            actual_clean = self.cleaner.clean_word(actual)
            target_clean = self.cleaner.clean_word(target)

            w.word = actual_clean
            w.word_actual = actual_clean
            w.word_target = target_clean



from acqdiv.parsers.metadata.parser import MetadataParser


class CMDIParser(MetadataParser):

    namespace = ('{http://www.clarin.eu/cmd/1/profiles/clarin.eu:'
                 'cr1:p_1407745712035}')

    def __init__(self, path):
        super().__init__(path)
        self.metadata["participants"] = self.get_participants()
        self.metadata["session"] = self.get_session_data()
        self.metadata["media"] = self.get_media_data()

    def get_participants(self):
        participants = []
        for actor in self.root.Actors.getchildren():
            participant = {}

            for actor_node in actor.getchildren():
                actor_tag = actor_node.tag.replace(self.namespace, '')

                if actor_tag == 'Actor_Languages':
                    langlist = []
                    for lang in actor_node.iterfind(
                            self.namespace + 'Actor_Language'):
                        langlist.append(lang.Name.text)
                    if len(langlist) != 0:
                        participant['languages'] = ','.join(langlist)
                else:
                    # skip empty fields in the IMDI
                    if not str(actor_node.text).strip() == "":
                        participant[actor_tag.lower()] = \
                            str(actor_node.text).strip()

            if len(participant):
                participants.append(participant)

        return participants

    def get_session_data(self):
        session = {
            'id': self.metadata['__attrs__']['Cname'],
            'date': self.root.Date.text
        }

        return session

    def get_media_data(self):
        media = {
            'mediafile': [],
            'writtenresource': []
        }

        return media



import re

from acqdiv.parsers.chat.readers.reader import CHATReader


class YucatecReader(CHATReader):

    @staticmethod
    def get_utterance_words(utterance):
        """Get utterance words.

        Also treats `&` as a word separator.
        """
        if utterance:
            return re.split(r'\s+|&', utterance)
        else:
            return []

    def get_morph_tier(self):
        return self.record.dependent_tiers.get('xmor', '')

    @classmethod
    def get_morpheme_words(cls, morph_tier):
        """Get morpheme words.

        Words are separated by blank spaces as well as & and + in the case of
        clitics.
        """
        if morph_tier:
            return re.split(r'\s+|&|\+', morph_tier)
        else:
            return []

    @staticmethod
    def iter_morphemes(word):
        """Iter morphemes of a word.

        Morphemes are separated by '#' (prefixes), ':' (suffixes) and '-'
        (unstructured morpheme).

        Glosses and POS tags may have sub glosses and POS tags with are
        separated from the main gloss/POS tag by ':'. This use can be
        distinguished from the morpheme-separating use by checking the strings
        to the left and right of the ':' – when they consist of nothing but
        uppercase letters and digits, they are sub glosses or POS tags;
        otherwise they belong to different morphemes.

        Morphemes containing a '|' are structured, whereas those not containing
        one are unstructured. The part to the left is either the gloss or the
        POS tag, while the part to the right is the segment. A word can be
        completely or partially unstructured. The rules are as follows:
        - Word completely unstructured: The morphemes are separated by dashes.
          The segment is filled, while the gloss and POS tag remain unfilled.
        - Prefixes:
            - structured: segment is right block, gloss is left block, POS tag
              is 'pfx'
            - unstructured (NOT ATTESTED!): segment is unfilled, morpheme is
              gloss, POS tag is 'pfx'
        - Stem:
            - structured: segment is right block, left block is POS tag if it
              is in the list `stem_poses`, otherwise it is a gloss
            - unstructured: if morpheme only consists of uppercase letters and
              digits, it is a gloss, otherwise it is a segment
        - Suffixes:
            - structured: segment is right block, gloss is left block, POS tag
              is 'sfx'
            - unstructured: segment is unfilled, morpheme is gloss, POS tag is
              'sfx'
        """
        # untranscribed word
        if word == 'xxx':
            yield '', '', ''
        # completely unstructured word
        elif not re.search(r'[:|]', word) and '-' in word:
            for morpheme in word.split('-'):
                seg = morpheme
                gloss = ''
                pos = ''
                yield seg, gloss, pos
        # fully or partially structured word
        else:
            morph_regex = re.compile(
                r'(?P<prefixes>.*#)?'
                r'((?P<stemleft>[0-9A-Z.:]+)\|-?)?(?P<stemright>[^:\-]+)'
                r'(?P<suffixes>[:\-].+)?')

            match = morph_regex.fullmatch(word)

            # ----- prefixes -----

            # if there are prefixes
            if match.group('prefixes'):
                prefix_string = match.group('prefixes').rstrip('#')
                # iter prefixes
                for pfx in prefix_string.split('#'):
                    pfx_structured = re.search(r'(.*)\|(.+)', pfx)
                    # structured prefixes
                    if pfx_structured is not None:
                        seg = pfx_structured.group(2)
                        gloss = pfx_structured.group(1)
                        pos = 'pfx'
                    # unstructured prefixes, tends to be gloss
                    else:
                        seg = ''
                        gloss = pfx
                        pos = 'pfx'
                    yield seg, gloss, pos

            # ----- stem -----

            # structured stems
            if match.group('stemleft'):
                seg = match.group('stemright')
                stem_left = match.group('stemleft')
                # stems with lexical meaning with the following POS tags
                # TODO: recheck this list
                stem_poses = {
                    '3PRON', 'ADJ', 'ADV', 'AUX',
                    'CLFR', 'CLFR.INAN', 'CLFR:INAN',
                    'CONJ', 'DEICT', 'DEM', 'DET', 'INT', 'INTERJ',
                    'N', 'N.PROP', 'N:PROP',
                    'NUM', 'PREP', 'PTL', 'QUANT', 'S',
                    'V', 'V.AUX', 'V:AUX', 'VI', 'V.INTRANS', 'V:INTRANS',
                    'VT', 'V.TRANS', 'V:TRANS'}
                if stem_left in stem_poses:
                    gloss = ''
                    pos = stem_left
                else:
                    gloss = stem_left
                    pos = ''
            # unstructured stems
            else:
                stem_right = match.group('stemright')
                if re.fullmatch(r'[A-Z0-9]+', stem_right):
                    seg = ''
                    gloss = stem_right
                    pos = ''
                else:
                    seg = stem_right
                    gloss = ''
                    pos = ''

            yield seg, gloss, pos

            # ----- suffixes -----

            # if there are suffixes
            if match.group('suffixes'):
                suffix_string = match.group('suffixes').lstrip(':').lstrip('-')
                # iter suffixes
                for sfx in re.split(r'(?<![A-Z1-9]):|(?<!\|)-', suffix_string):
                    sfx_structured = re.search(r'(.*)\|-?(.+)', sfx)
                    # structured suffixes
                    if sfx_structured is not None:
                        seg = sfx_structured.group(2)
                        gloss = sfx_structured.group(1)
                        pos = 'sfx'
                    # unstructured suffixes: tends to be gloss
                    else:
                        seg = ''
                        gloss = sfx
                        pos = 'sfx'
                    yield seg, gloss, pos

    def get_translation(self):
        return self.record.dependent_tiers.get('xspn', '')

    @classmethod
    def get_segments(cls, seg_word):
        return [seg for seg, _, _ in cls.iter_morphemes(seg_word)]

    @classmethod
    def get_glosses(cls, gloss_word):
        return [gloss for _, gloss, _ in cls.iter_morphemes(gloss_word)]

    @classmethod
    def get_poses(cls, pos_word):
        return [pos for _, _, pos in cls.iter_morphemes(pos_word)]





def align_words_morphemes(utt):
    """Align words and morphemes of an utterance.

    Sets the word of the morpheme that it belongs to when there are
    no misalignments. Also, copies the POS and POS UD to the word.

    Args:
        utt (acqdiv.model.utterance.Utterance): The utterance.
    """
    link_to_word = len(utt.morphemes) == len(utt.words)
    for i, mword in enumerate(utt.morphemes):
        for morpheme in mword:
            if link_to_word:
                morpheme.word = utt.words[i]
                if morpheme.pos not in ['sfx', 'pfx']:
                    utt.words[i].pos = morpheme.pos
                    utt.words[i].pos_ud = morpheme.pos_ud


def fix_misalignments(entities):
    """Fix misalignments.

    Correct mismatches in the number of entities.

    Example:
        fix_misalignments([['seg1', 'seg2'], ['gloss1'], ['pos1', 'pos2']])
        => [['seg1', 'seg2'], ['', ''], ['pos1', 'pos2']]

    Args:
        entities (List[List[Any]]): The entities.

    Returns:
        List[List[Any]]: The aligned entities.
    """
    n = _get_number_of_entities(entities)
    _adjust_misaligned_entities(n, entities)

    return entities


def _get_number_of_entities(entities):
    """Get number of entities.

    Args:
        entities (List[List[Any]]]): The entities

    These entities may be words or morphemes. The number of entities is
    calculated based on:
     - The order of the lists of entities.
     - Whether the lists are empty or not.

     Returns: int
    """
    for entities_set in entities:
        if len(entities_set):
            return len(entities_set)

    return 0


def _adjust_misaligned_entities(n, entities):
    """Adjust misaligned entities.

    Lists diverging from `n` in terms of the number of entities that
    they contain will be adjusted in such a way that they have
    the same number of entities which are empty.

    Args:
        n (int): The number of entities to be considered.
        entities (List[List[Any]]]): The entities.
    """
    for i, entity_list in enumerate(entities):
        if n != len(entity_list):
            entities[i] = n * ['']





class Record:

    def __init__(self):
        self.tiers = {}

    def get(self, field, not_exists=''):
        return self.tiers.get(field, not_exists)

    def __getitem__(self, field):
        return self.tiers.get(field, '')

    def __setitem__(self, key, value):
        self.tiers[key] = value

    def __iter__(self):
        for tier in self.tiers:
            yield tier

    def __contains__(self, field):
        return field in self.tiers



import re

from acqdiv.util.csvparser import parse_csv
from acqdiv.util.path import get_full_path


class TuatschinGlossMapper:

    gloss_dict = parse_csv(get_full_path(
        'parsers/corpora/main/tuatschin/resources/gloss.csv'))

    @classmethod
    def map(cls, gloss):
        gloss = cls.clean_gloss(gloss)

        if gloss:
            # replace person/number combinations first
            pnum_regex = re.compile(r'([0123])\.(Sing)')
            gloss = pnum_regex.sub(r'\1SG', gloss)
            pnum_regex = re.compile(r'([0123])\.(Plur)')
            gloss = pnum_regex.sub(r'\1PL', gloss)

            parts = []
            is_null = False
            for part in gloss.split('.'):
                if re.search(r'[0123](SG|PL)', part):
                    parts.append(part)
                else:
                    if part in cls.gloss_dict:
                        part = cls.gloss_dict[part]

                        if part != '???':
                            parts.append(part)
                        else:
                            is_null = True
                            break
                    else:
                        is_null = True
                        break

            if is_null:
                gloss = ''
            else:
                gloss = '.'.join(parts)

        else:
            gloss = ''

        return gloss

    @classmethod
    def clean_gloss(cls, gloss):
        for cleaning_method in [
            cls.remove_pos
        ]:
            gloss = cleaning_method(gloss)
        return gloss

    @staticmethod
    def remove_pos(gloss):
        """Remove the POS tag.

        Morpho-syntactic annotations start with the POS tag:
        [POS].[SUB-GlOSS1].[SUB-GLOSS2]

        Example:
            ADJ.Fem.Sing => Fem.Sing
        """
        regex = re.compile(r'^[^.]+\.')
        gloss = regex.sub('', gloss)

        return gloss



from typing import List, Optional

from acqdiv.model.speaker import Speaker
from acqdiv.model.word import Word
from acqdiv.model.morpheme import Morpheme


class Utterance:
    """Data class representing an utterance.

    source_id (str): The ID of the utterance.
    speaker_label (str): The label of the speaker that makes the utterance.
    addressee (str): The label of the speaker whom the utterance is
        addressed at.
    utterance_raw (str): The original utterance.
    utterance (str): The cleaned utterance.
    actual_utterance (str): The actual utterance.
    target_utterance (str): The target utterance.
    translation (str): The translation of the utterance.
    morpheme_raw (str): The original morpheme tier.
    morpheme (str): The cleaned morpheme tier.
    gloss_raw (str): The original gloss tier.
    gloss (str): The cleaned gloss tier.
    pos_raw (str): The original POS tag tier.
    pos (str): The cleaned POS tag tier.
    sentence_type: The type of the utterance.
    childdirected (bool): Whether the utterance is child directed.
    start_raw (str): The original start time of the utterance.
    start (str): The cleaned start time of the utterance.
    end_raw (str): The original end time of the utterance.
    end (str): The cleaned end time of the utterance.
    comment (str): Any comments about the utterance.
    warning (str): Warnings concering the utterance.
    words (List[Word]): The words of the utterance.
    morphemes (List[Morpheme]]): The morphemes of the utterance.
    """

    source_id: str
    speaker: Optional[Speaker]
    addressee: Optional[Speaker]
    utterance_raw: str
    utterance: str
    actual_utterance: str
    target_utterance: str
    translation: str
    morpheme_raw: str
    morpheme: str
    gloss_raw: str
    gloss: str
    pos_raw: str
    pos: str
    sentence_type: str
    childdirected: str
    start_raw = str
    start = str
    end_raw: str
    end: str
    comment: str
    warning: str
    words: List[Word]
    morphemes: List[List[Morpheme]]

    def __init__(self):
        """Initialize the variables representing an utterance."""
        self.source_id = ''
        self.speaker = None
        self.addressee = None
        self.utterance_raw = ''
        self.utterance = ''
        self.actual_utterance = ''
        self.target_utterance = ''
        self.translation = ''
        self.morpheme_raw = ''
        self.morpheme = ''
        self.gloss_raw = ''
        self.gloss = ''
        self.pos_raw = ''
        self.pos = ''
        self.sentence_type = ''
        self.childdirected = ''
        self.start_raw = ''
        self.start = ''
        self.end_raw = ''
        self.end = ''
        self.comment = ''
        self.warning = ''
        self.words = []
        self.morphemes = []




from acqdiv.parsers.chat.cleaners.cleaner import CHATCleaner
from acqdiv.parsers.corpora.main.japanese_miipro.cleaner import \
    JapaneseMiiProCleaner
from acqdiv.parsers.chat.cleaners.utterance_cleaner \
    import CHATUtteranceCleaner
from acqdiv.parsers.corpora.main.japanese_miyata.gloss_mapper \
    import JapaneseMiyataGlossMapper as GMp
from acqdiv.parsers.corpora.main.japanese_miyata.pos_mapper \
    import JapaneseMiyataPOSMapper as PMp


class JapaneseMiyataCleaner(CHATCleaner):

    # ---------- morphology tier cleaning ----------

    @classmethod
    def remove_dloc(cls, morph_tier):
        """Remove dloc|dloc=DISLOC words on the morphology tier.

        dloc|dloc=DISLOC stands for `„` on the utterance.
        """
        morph_tier = morph_tier.replace('dloc|dloc=DISLOC', '')
        return CHATUtteranceCleaner.remove_redundant_whitespaces(morph_tier)

    @classmethod
    def clean_morph_tier(cls, morph_tier):
        morph_tier = CHATUtteranceCleaner.remove_terminator(morph_tier)
        morph_tier = cls.remove_dloc(morph_tier)
        return morph_tier

    # ---------- utterance cross clean ----------

    @classmethod
    def utterance_cross_clean(
            cls, raw_utt, actual_utt, target_utt,
            seg_tier, gloss_tier, pos_tier):

        return JapaneseMiiProCleaner.utterance_cross_clean(
            raw_utt,
            actual_utt,
            target_utt,
            seg_tier,
            gloss_tier,
            pos_tier)

    # ---------- speaker metadata cleaning ----------

    @classmethod
    def clean_speaker_metadata(
            cls, session_filename, label, name, role,
            age, gender, language, birth_date, target_child):
        """Correct label, role and name of speaker."""

        _, tc_name = target_child
        name = cls.correct_speaker_name(tc_name, name, label)
        role = cls.correct_role(role, session_filename, label)

        return label, name, role, age, gender, language, birth_date

    @classmethod
    def correct_role(cls, role, session_filename, speaker_label):
        if session_filename.startswith('Tai') and speaker_label == 'TAI':
            return 'Target_Child'

        return role

    @classmethod
    def correct_speaker_name(cls, tc_name, name, label):
        if tc_name == 'Akifumi' and name == 'Okaasan':
            return 'Mother_of_AKI'
        elif tc_name == 'Ryookun':
            if name == 'Okaasan':
                return 'Mother_of_RYO'
            elif name == 'Papa':
                return 'Father_of_RYO'

        if label == 'REE':
            return 'Ree'

        if label == 'TOS':
            return 'Miyatanopapa'

        return name

    # ---------- session metadata cleaning ----------

    @classmethod
    def clean_session_metadata(cls, session_filename, date, media_filename):
        """Correct the session date."""
        return date, media_filename

    # ---------- morpheme cleaning ----------

    @classmethod
    def clean_gloss(cls, gloss):
        return GMp.map(gloss)

    @classmethod
    def clean_pos(cls, pos):
        return PMp.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return PMp.map(pos_ud, ud=True)




class UniqueSpeaker:
    """Data class representing a unique speaker in a corpus.

    code (str): The label of the speaker.
    name (str): The name of the speaker.
    gender_raw (str): The original gender of the speaker.
    gender (str): The cleaned gender of the speaker.
    birth_date (str): The birth date of the speaker.
    """

    code: str
    name: str
    gender_raw: str
    gender: str
    birth_date: str

    def __init__(self):
        self.code = ''
        self.name = ''
        self.gender_raw = ''
        self.gender = ''
        self.birth_date = ''



import re

from acqdiv.util.csvparser import parse_csv
from acqdiv.util.path import get_full_path


class NungonGlossMapper:

    gloss_dict = parse_csv(get_full_path(
        'parsers/corpora/main/nungon/resources/gloss.csv'))

    @classmethod
    def map(cls, gloss):
        gloss = cls.clean_gloss(gloss)
        return cls.gloss_dict.get(gloss, '')

    @classmethod
    def clean_gloss(cls, gloss):
        for cleaning_method in [
                cls.remove_question_mark,
                cls.replace_slash,
                cls.replace_plus
        ]:
            gloss = cleaning_method(gloss)
        return gloss

    @staticmethod
    def remove_question_mark(morpheme):
        """Remove the question mark in the morpheme.

        Question marks might code insecure annotations. They are prefixed to
        the morpheme.
        """
        return morpheme.lstrip('?')

    @staticmethod
    def replace_slash(gloss):
        """Replace the slash by a dot between numbers."""
        return re.sub(r'(\d)/(\d)', r'\1.\2', gloss)

    @staticmethod
    def replace_plus(gloss):
        """Replace the plus by a dot."""
        return gloss.replace('+', '.')




def clean(session):
    """Correct multiple target children in session.

    Args:
        session (acqdiv.model.session.Session): The session.
    """
    tcs = [speaker for speaker in session.speakers
           if speaker.macro_role == 'Target_Child']

    if len(tcs) > 1:
        label = session.source_id[:3]
        for tc in tcs:
            if tc.code != label:
                tc.role = 'Child'
                tc.macro_role = 'Child'



import re


class SentenceTypeExtractor:
    """Methods for inferring the sentence type of a CHAT utterance."""

    @classmethod
    def get_sentence_type(cls, utterance):
        """Get the sentence type of the utterance.

        Args:
            utterance (str): The utterance.

        Returns:
            str: The sentence type.
        """
        terminator = cls.get_utterance_terminator(utterance)
        sentence_type = cls.terminator2sentence_type(terminator)
        return sentence_type

    @staticmethod
    def get_utterance_terminator(utterance):
        terminator_regex = re.compile(r'([+/.!?"]*[!?.])(?=(\s*\[\+|\s*$))')
        match = terminator_regex.search(utterance)
        if match:
            return match.group(1)
        else:
            return ''

    @staticmethod
    def terminator2sentence_type(terminator):
        """Map utterance terminator to sentence type.

        Returns:
            str: The sentence type.
        """
        mapping = {'.': 'default',
                   '?': 'question',
                   '!': 'exclamation',
                   '+.': 'transcription break',
                   '+...': 'trail off',
                   '+..?': 'trail off of question',
                   '+!?': 'question with exclamation',
                   '+/.': 'interruption',
                   '+/?': 'interruption of a question',
                   '+//.': 'self-interruption',
                   '+//?': 'self-interrupted question',
                   '+"/.': 'quotation follows',
                   '+".': 'quotation precedes'}

        return mapping.get(terminator, '')





class IMDICleaner:

    null_values = {'Unknown', 'Unspecified', 'None', 'Unidentified'}

    @classmethod
    def unify_unknowns(cls, value):
        if value in cls.null_values:
            return ''
        else:
            return value

    @classmethod
    def clean_date(cls, date):
        return cls.unify_unknowns(date)

    @classmethod
    def clean_name(cls, name):
        return cls.unify_unknowns(name)

    @classmethod
    def clean_gender(cls, gender):
        return cls.unify_unknowns(gender)

    @classmethod
    def clean_label(cls, label):
        return cls.unify_unknowns(label)



from setuptools import setup, find_packages


setup(
    name='acqdiv',
    version='0.2.0',
    description='Pipeline for the ACQDIV database',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/acqdiv/acqdiv',
    author='ACQDIV developer team',
    author_email='steven.moran@uzh.ch',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Text Processing :: Linguistic',
        'License :: Free For Educational Use',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    keywords='data linguistics',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=[
        'sqlalchemy',
        'lxml',
        'tqdm',
        'python-dateutil'
    ],
    extras_require={
        'dev': ['pandas', 'numpy', 'tox'],
        'test':  ['pytest'],
    },
    entry_points={
        'console_scripts': ['acqdiv=acqdiv.__main__:main'],
    },
    project_urls={
        'Bug Reports': 'https://github.com/acqdiv/acqdiv/issues',
        'Source': 'https://github.com/acqdiv/acqdiv',
    }
)



import re

from acqdiv.parsers.chat.cleaners.cleaner import CHATCleaner
from acqdiv.parsers.chat.cleaners.utterance_cleaner \
    import CHATUtteranceCleaner
from acqdiv.parsers.corpora.main.yucatec.gloss_mapper \
    import YucatecGlossMapper
from acqdiv.parsers.corpora.main.yucatec.pos_mapper \
    import YucatecPOSMapper


class YucatecCleaner(CHATCleaner):

    # ---------- utterance cleaning ----------

    # TODO: removing dashes in utterance?

    @classmethod
    def remove_terminator(cls, utterance):
        """Remove utterance terminator.

        Also removes the colon and the dash.
        """
        utterance = CHATUtteranceCleaner.remove_terminator(utterance)
        return utterance.rstrip('-').rstrip(':')

    # ---------- morphology tier cleaning ----------

    @classmethod
    def remove_double_hashes(cls, morph_tier):
        """Remove ## from the morphology tier."""
        morph_tier = re.sub(r'(^| )##( |$)', r'\1\2', morph_tier)
        return CHATUtteranceCleaner.remove_redundant_whitespaces(morph_tier)

    @classmethod
    def clean_morph_tier(cls, morph_tier):
        for cleaning_method in [
                cls.remove_terminator, cls.remove_double_hashes]:
            morph_tier = cleaning_method(morph_tier)

        return morph_tier

    # ---------- morpheme word cleaning ----------

    @staticmethod
    def correct_hyphens(morpheme_word):
        """Replace faulty hyphens by the pipe in the morpheme word.

        It is only attested in suffixes, not in prefixes.
        """
        return re.sub(r'(:[A-Z0-9]+)-(?=[a-záéíóúʔ]+)', r'\1|', morpheme_word)

    @staticmethod
    def remove_colon(morpheme_word):
        """Remove leading and trailing colons.

        Note:
            In some cases, the colons are correct, but a whitespace to the
            preceding or following morpheme was erroneously inserted. This
            case is not handled as the inference rules would be quite complex.
        """
        return morpheme_word.strip(':')

    @staticmethod
    def remove_dash(word_morpheme):
        """Remove leading and trailing dash.

        Note:
            In some cases, the dashes are correct, but a whitespace to the
            preceding or following morpheme was erroneously inserted. This
            case is not handled as the inference rules would be quite complex.
        """
        return word_morpheme.strip('-')

    @staticmethod
    def remove_colon_dash(word_morpheme):
        """Remove trailing colon and dash.

        Note:
            In some cases, they are correct, but a whitespace to the
            preceding or following morpheme was erroneously inserted. This
            case is not handled as the inference rules would be quite complex.
        """
        return word_morpheme.rstrip('-').rstrip(':')

    @classmethod
    def clean_morpheme_word(cls, morpheme_word):
        for cleaning_method in [
                cls.correct_hyphens, cls.remove_colon, cls.remove_dash,
                cls.remove_colon_dash]:
            morpheme_word = cleaning_method(morpheme_word)
        return morpheme_word

    # ---------- morpheme cleaning ----------

    @classmethod
    def clean_gloss(cls, gloss):
        return YucatecGlossMapper.map(gloss)

    @classmethod
    def clean_pos(cls, pos):
        return YucatecPOSMapper.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return YucatecPOSMapper.map(pos_ud, ud=True)



from acqdiv.parsers.corpus_parser import CorpusParser
from acqdiv.parsers.corpora.main.english.session_parser \
    import EnglishManchester1SessionParser


class EnglishCorpusParser(CorpusParser):

    def get_session_parser(self, session_path):
        return EnglishManchester1SessionParser(session_path)



import re

from acqdiv.parsers.toolbox.cleaners.cleaner import ToolboxCleaner
from acqdiv.parsers.corpora.main.ku_waru.pos_mapper import KuWaruPOSMapper
from acqdiv.parsers.corpora.main.ku_waru.gloss_mapper \
    import KuWaruGlossMapper


class KuWaruCleaner(ToolboxCleaner):

    @classmethod
    def unify_unknown(cls, utterance):
        return utterance.replace('***', '???')

    @classmethod
    def remove_punctuation(cls, utterance):
        utterance = re.sub(r'[?.,)(]', '', utterance)
        return cls.remove_redundant_whitespaces(utterance)

    @classmethod
    def clean_utterance(cls, utterance):
        utterance = cls.remove_punctuation(utterance)
        utterance = cls.unify_unknown(utterance)
        return utterance

    @classmethod
    def clean_gloss(cls, gloss):
        return KuWaruGlossMapper.map(gloss)

    @classmethod
    def clean_pos(cls, pos):
        return KuWaruPOSMapper.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return KuWaruPOSMapper.map(pos_ud, ud=True)



from acqdiv.parsers.toolbox.cleaners.cleaner import ToolboxCleaner


class DeneCleaner(ToolboxCleaner):

    @classmethod
    def unify_unknown(cls, utterance):
        return utterance.replace('xxx', '???')





class IndonesianPOSMapper:

    @classmethod
    def map(cls, pos, ud=False):
        """Infer POS from the gloss.

        There is no POS tier in Indonesian, but the macro categories
        `sfx`, `pfx`, `stem` can be inferred from the gloss.
        """
        if ud:
            return ''
        else:
            if pos.startswith('-'):
                return 'sfx'
            elif pos.endswith('-'):
                return 'pfx'
            elif pos in ['', '???']:
                return pos
            else:
                pos = cls.substitute(pos)
                if pos:
                    return pos
                else:
                    return 'stem'

    @staticmethod
    def substitute(pos):
        labels = {'BO', 'DAH', 'DENG', 'DING', 'EH', 'NAH'}
        if pos in labels:
            return 'PTCL'
        else:
            return ''



import re

from acqdiv.parsers.chat.readers.reader import CHATReader


class SesothoReader(CHATReader):
    """Class to implement Sesotho Reading methods.

    Normally the methods get_utterance(), get_actual_utterance() and
    get_target_utterance() extract the utterance from the mainline.
    In Sesotho there are many misalignments between the
    mainline-utterance and the morpheme-tiers. So instead the
    utterance (for all three methods) is created by joining morphemes to
    words and those to the utterance.
    """

    def __init__(self, session_file):
        super().__init__(session_file)
        self._passed_stem = False

    def _join_morph_to_utt(self):
        """Create the utterance by joining the morphemes to words.

        Morphemes are separated by hyphens. The hyphens are removed.
        """
        seg_tier = self.get_seg_tier()
        seg_words = seg_tier.split(' ')
        utt_words = []

        for w in seg_words:
            w = ''.join(w.split('-'))
            utt_words.append(w)

        return ' '.join(utt_words)

    def get_actual_utterance(self):
        """Get the actual utterance, created by joined morphemes."""
        return self._join_morph_to_utt()

    def get_target_utterance(self):
        """Get the target utterance, created by joined morphemes."""
        return self._join_morph_to_utt()

    def get_seg_tier(self):
        """Extract the segments tier and do cross-cleaning.

        Cross-cleaning has to be done from here for access to different
        tiers at the same time.
        """
        return self.record.dependent_tiers.get('gls', '')

    def get_gloss_tier(self):
        return self.record.dependent_tiers.get('xcod', '')

    def get_pos_tier(self):
        return self.record.dependent_tiers.get('xcod', '')

    def infer_pos(self, gloss, num_morphemes):
        """Infer the pos-tag for a gloss.

        Args:
            gloss: str, the gloss
            num_morphemes: int, the number of morphemes in the
                morpheme word
        Return:
            str, the pos-tag
        """
        if not gloss:
            return ''

        pos = ''
        # Check for prefixes and suffixes.
        if num_morphemes == 1 or (re.search(r'(v|id)\^|\(\d', gloss)
                                  or re.match(r'(aj$|nm$|ps\d+)', gloss)):
            self._passed_stem = True
            # Check for verbs: verbs have v^, one typo as s^.
            if re.search(r'[vs]\^', gloss):
                pos = 'v'

            # Check for nouns: nouns contains "(\d)" (default) or "ps/"
            elif re.search('\(\d+', gloss) or re.search('^ps/', gloss):
                pos = 'n'

            # Check for words with nominal concord.
            elif re.search(r'^(d|lr|obr|or|pn|ps|sr)\d+', gloss):
                pos_match = re.search('^(d|lr|obr|or|pn|ps|sr)\d+', gloss)
                pos = pos_match.group(1)

            # Check for particles: mostly without a precise gloss.
            elif re.search(
                    (r'^(aj|av|cd|cj|cm|ht|ij|loc|lr|ng|nm|obr|or|pr|q|sr'
                     r'|wh)$'),
                    gloss):
                pos = gloss

            # Check for free person markers.
            elif re.search(r'^sm\d+[sp]?$', gloss):
                pos = 'afx.detached'

            # Check for copulas.
            elif re.search(r'^cp|cp$', gloss):
                pos = 'cop'

            # Check for ideophones.
            elif re.search(r'id\^', gloss):
                pos = 'ideoph'

            # Check for meaningless and unclear words. Note that
            # "xxx" in the Sesotho coding tier is not the same as
            # CHAT "xxx" in the transcription tier - it does not
            # stand for words that could not be transcribed but for
            # words with unclear meaning.
            elif gloss == 'word' or gloss == 'xxx':
                pos = 'none'
            else:
                pos = '???'

        elif not self._passed_stem:
            pos = 'pfx'
        elif self._passed_stem:
            pos = 'sfx'

        return pos

    def iter_gloss_pos(self, gloss_word):
        """Iter glosses and poses of a gloss_word.

        This method only returns glosses and poses, but not segments,
        since segments are on a different tier. They are retrieved
        directly by get_segments().

        Glosses in a gloss_words (stems and affixes) are separated
        by hyphens.

        Args:
            gloss_word: str, the gloss of a word
        Returns:
            tuple: (gloss, pos).
        """
        if not gloss_word:
            yield ('', '')
        else:
            glosses = gloss_word.split('-')
            self._passed_stem = False
            for gloss in glosses:
                pos = self.infer_pos(gloss, len(glosses))
                yield (gloss, pos)

    @classmethod
    def get_segments(cls, seg_word):
        return seg_word.split('-')

    def get_glosses(self, gloss_word):
        return [gloss for gloss, _ in self.iter_gloss_pos(gloss_word)]

    def get_poses(self, pos_word):
        return [pos for _, pos in self.iter_gloss_pos(pos_word)]



import re


class ToolboxReader(object):
    """Methods for reading Toolbox."""

    # ---------- record ----------

    @classmethod
    def is_record(cls, rec_dict):
        """Is the record really a record or just metadata?"""
        for tier in rec_dict:
            content = rec_dict[tier]

            if content.startswith('@'):
                return False

        return True

    @staticmethod
    def get_rec_separator():
        return br'\\ref'

    # ---------- tiers ----------

    @classmethod
    def get_source_id(cls, rec):
        return rec.get('ref', '')

    @classmethod
    def get_speaker_label(cls, rec):
        return rec.get('ELANParticipant', '')

    @classmethod
    def get_addressee(cls, rec):
        return rec.get('add', '')

    @classmethod
    def get_start_raw(cls, rec):
        return rec.get('ELANBegin', '')

    @classmethod
    def get_end_raw(cls, rec):
        return rec.get('ELANEnd', '')

    @classmethod
    def get_actual_utterance(cls, rec):
        return rec.get('tx', '')

    @classmethod
    def get_target_utterance(cls, rec):
        return rec.get('target', '')

    @classmethod
    def get_sentence_type(cls, rec):
        """Get utterance type (aka sentence type) of an utterance.

        Possible values:
            - default (.)
            - question (?)
            - imperative or exclamation (!)

        Args:
            rec (acqdiv.parsers.toolbox.model.record.Record): The record.

        Returns:
            str: The sentence type.
        """
        utterance_raw = cls.get_actual_utterance(rec)
        match_punctuation = re.search('([.?!])$', utterance_raw)
        if match_punctuation is not None:
            if match_punctuation.group(1) == '.':
                return 'default'
            elif match_punctuation.group(1) == '?':
                return 'question'
            elif match_punctuation.group(1) == '!':
                return 'imperative'

        return ''

    @classmethod
    def get_childdirected(cls, rec):
        """Not coded per default.

        Args:
            rec (acqdiv.parsers.toolbox.model.record.Record): The record.
        """
        return ''

    @classmethod
    def get_translation(cls, rec):
        return rec.get('eng', '')

    @classmethod
    def get_comment(cls, rec):
        return rec.get('comment', '')

    @classmethod
    def get_seg_tier(cls, rec):
        return rec.get('mb', '')

    @classmethod
    def get_gloss_tier(cls, rec):
        return rec.get('ge', '')

    @classmethod
    def get_pos_tier(cls, rec):
        return rec.get('ps', '')

    @classmethod
    def get_lang_tier(cls, rec):
        return rec.get('lg', '')

    @classmethod
    def get_id_tier(cls, rec):
        return rec.get('lemma_id', '')

    # ---------- words ----------

    @classmethod
    def get_words(cls, utterance):
        return utterance.split()

    @classmethod
    def get_morpheme_words(cls, morpheme_tier):
        _word_boundary = re.compile(r'(?<![\-=\s])\s+(?![\-=\s])')

        if morpheme_tier:
            return re.split(_word_boundary, morpheme_tier)
        else:
            return []

    @classmethod
    def get_seg_words(cls, segment_tier):
        return cls.get_morpheme_words(segment_tier)

    @classmethod
    def get_gloss_words(cls, gloss_tier):
        return cls.get_morpheme_words(gloss_tier)

    @classmethod
    def get_pos_words(cls, pos_tier):
        return cls.get_morpheme_words(pos_tier)

    @classmethod
    def get_lang_words(cls, morpheme_lang_tier):
        return cls.get_morpheme_words(morpheme_lang_tier)

    @classmethod
    def get_id_words(cls, id_tier):
        return cls.get_morpheme_words(id_tier)

    # ---------- morphemes ----------

    @classmethod
    def get_morphemes(cls, morpheme_word):
        if morpheme_word:
            return morpheme_word.split()
        else:
            return []

    @classmethod
    def get_segs(cls, segment_word):
        return cls.get_morphemes(segment_word)

    @classmethod
    def get_glosses(cls, gloss_word):
        return cls.get_morphemes(gloss_word)

    @classmethod
    def get_poses(cls, pos_word):
        return cls.get_morphemes(pos_word)

    @classmethod
    def get_langs(cls, morpheme_lang_word):
        return cls.get_morphemes(morpheme_lang_word)

    @classmethod
    def get_ids(cls, id_word):
        return cls.get_morphemes(id_word)

    @staticmethod
    def get_morpheme_type():
        return 'target'

    @staticmethod
    def get_main_morpheme():
        return 'gloss'



import re

from acqdiv.parsers.chat.readers.reader import CHATReader


class NungonReader(CHATReader):

    # ---------- morphology tier ----------

    def get_seg_tier(self):
        """Get the segment tier.

        The name of the segment tier can be either 'xgls' or 'gls'. The former
        is more common.
        """
        for seg_tier_name in ['xgls', 'gls']:
            if seg_tier_name in self.record.dependent_tiers:
                return self.record.dependent_tiers[seg_tier_name]
        return ''

    def get_gloss_tier(self):
        return self.record.dependent_tiers.get('xcod', '')

    def get_pos_tier(self):
        return self.record.dependent_tiers.get('xcod', '')

    # ---------- morpheme words ----------

    @classmethod
    def get_morpheme_words(cls, morph_tier):
        """Get the morpheme words of the morphology tier.

        Morpheme words are normally separated by whitespaces. Clitics are
        separated by '=' and are split off in parsing as they correspond
        to independent words in the utterance.
        """
        if morph_tier:
            return re.split(r'\s+|=', morph_tier)
        else:
            return []

    # ---------- morphemes ----------

    @classmethod
    def get_segments(cls, seg_word):
        """Segments are separated by dashes."""
        if seg_word:
            return seg_word.split('-')
        else:
            return []

    @staticmethod
    def iter_gloss_pos(gloss_pos_word):
        """Iter glosses and POS tags of a word.

        Morphemes are separated by dashes ('-'). Stems have a POS tag which is
        prefixed to the gloss by a caret ('^'). Prefixes and suffixes have only
        explicit glosses. They receive the POS tag 'pfx' and 'sfx',
        respectively. Some words do not have a stem marker in which case all
        morphemes get only glosses but no POS tags assigned.

        Yields:
            tuple: (gloss, POS tag)
        """
        morphemes = gloss_pos_word.split('-')

        # check if there is a stem marker
        if '^' not in gloss_pos_word:
            for gloss in morphemes:
                yield gloss, ''
        else:
            stem_passed = False
            for morpheme in morphemes:
                # check if it is the stem
                if '^' in morpheme:
                    # match POS up to the last ^ (in case there are several ^)
                    match = re.search(r'(.*)\^(.*)', morpheme)
                    gloss = match.group(2)
                    pos = match.group(1)
                    stem_passed = True
                else:
                    gloss = morpheme
                    if stem_passed:
                        pos = 'sfx'
                    else:
                        pos = 'pfx'

                yield gloss, pos

    @classmethod
    def get_glosses(cls, gloss_word):
        return [gloss for gloss, _ in cls.iter_gloss_pos(gloss_word)]

    @classmethod
    def get_poses(cls, pos_word):
        return [pos for _, pos in cls.iter_gloss_pos(pos_word)]

    @classmethod
    def get_morpheme_language(cls, seg, gloss, pos):
        """Get the morpheme language.

        The morpheme language is coded in the POS tag. If it starts with 'eng',
        the morpheme is English, if it starts with 'tp', the morpheme is Tok
        Pisin.
        """
        if pos.startswith('eng'):
            return 'English'
        elif pos.startswith('tp'):
            return 'Tok Pisin'
        else:
            return 'Nungon'



import re


class ToolboxMorphemeCleaner:

    @classmethod
    def clean(cls, morpheme):
        morpheme = cls.remove_morpheme_delimiters(morpheme)
        morpheme = cls.null_unknown(morpheme)

        return morpheme

    @staticmethod
    def remove_morpheme_delimiters(morpheme):
        """Remove morpheme delimiters.

        Morpheme delimiters are `-` and `=`.
        """
        return morpheme.strip('-').strip('=')

    @staticmethod
    def null_unknown(morpheme):
        """Return empty string for unknown values.

        Unknown values: ***, ???
        """
        return re.sub(r'\*{3}|\?{3}', '', morpheme)



import re

from acqdiv.parsers.chat.cleaners.cleaner import CHATCleaner
from acqdiv.parsers.corpora.main.cree.gloss_mapper \
    import CreeGlossMapper
from acqdiv.parsers.corpora.main.cree.pos_mapper \
    import CreePOSMapper


class CreeCleaner(CHATCleaner):

    @staticmethod
    def correct_name(name):
        if name == 'A1':
            return 'Ani'

        return name

    @staticmethod
    def correct_speaker_label(speaker_label, name):
        if name == 'Adult':
            return 'ADU'
        elif name == 'A1':
            return 'CHI'

        return speaker_label

    @staticmethod
    def correct_role(speaker_label, role):
        if speaker_label == 'CH2':
            return 'Child'

        return role

    @classmethod
    def clean_speaker_metadata(
            cls, session_filename, label, name, role,
            age, gender, language, birth_date, target_child):
        """Correct label, role and name of speaker."""

        label = cls.correct_speaker_label(label, name)
        name = cls.correct_name(name)
        role = cls.correct_role(label, role)

        return label, name, role, age, gender, language, birth_date

    @classmethod
    def clean_record_speaker_label(cls, session_filename, speaker_label):
        child = ['p12', 'p10', 'p21', 'p33', 'p0']
        adult = ['p13', 'p11', 'p22', 'p2', 'p34']

        if speaker_label in child:
            return 'CHI'

        elif speaker_label in adult:
            return 'ADU'

        if (session_filename == '20-A1-2006-10-03.cha'
                and speaker_label == 'p1'):
            return 'CHI'

        if (session_filename == '24-A1-2007-03-07.cha'
                and speaker_label == 'p1'):
            return 'ADU'

        return speaker_label

    # ---------- utterance cleaning ----------

    @staticmethod
    def remove_angle_brackets(utterance):
        """Remove the small angle brackets.

        The angle brackets are smaller than the standard angle brackets: ‹›
        (vs. <>). They occur around the utterance, but the closing bracket
        occurs before the utterance terminator. In CHAT, they are used for
        marking special alignment with the %pho tier.
        """
        return utterance.replace('‹', '').replace('›', '')

    @classmethod
    def clean_utterance(cls, utterance):
        utterance = cls.remove_angle_brackets(utterance)
        return super().clean_utterance(utterance)

    # ---------- word cleaning ----------

    @staticmethod
    def remove_morph_separators(word):
        """Remove morpheme separators in a word.

        An underscore is used as a morpheme separator (e.g. 'giddy_up').
        """
        morph_sep_regex = re.compile(r'(\S+?)_(\S+?)')
        return morph_sep_regex.sub(r'\1\2', word)

    @staticmethod
    def replace_zero(word):
        """Replace zéro morphemes in a word.

        'zéro' stands for zero morphemes and is replaced by 'Ø'.
        """
        return word.replace('zéro', 'Ø')

    @staticmethod
    def replace_morpheme_separator(word):
        """Replace morpheme separators in a word.

        Morphemes are separated by a tilde.
        """
        return word.replace('~', '')

    @classmethod
    def clean_word(cls, word):
        word = super().clean_word(word)
        for cleaning_method in [cls.remove_morph_separators, cls.replace_zero,
                                cls.replace_morpheme_separator]:
            word = cleaning_method(word)

        return word

    # ---------- morphology tier cleaning ----------

    @staticmethod
    def remove_square_brackets(morph_tier):
        """Remove redundant square brackets around morphology tiers.

        Morphology tiers have square brackets at their edges which can be
        removed. It is unclear what their purpose is.
        """
        return morph_tier.lstrip('[').rstrip(']')

    @classmethod
    def null_untranscribed_morph_tier(cls, morph_tier):
        """Null untranscribed morphology tier.

        Note:
            Nulling means here, an empty string is returned.
        """
        if morph_tier == '*':
            return ''

        return morph_tier

    @classmethod
    def clean_morph_tier(cls, morph_tier):
        morph_tier = cls.remove_square_brackets(morph_tier)
        # morph_tier = cls.null_untranscribed_morph_tier(morph_tier)
        return morph_tier

    # ---------- cross cleaning ----------

    @staticmethod
    def replace_eng(gloss_tier, utterance):
        """Replace the 'Eng' glosses by the actual words in the gloss tier.

        Returns:
            str: The gloss tier with all its 'Eng' glosses replaced.
        """
        gloss_words = gloss_tier.split(' ')
        utterance_words = utterance.split(' ')

        # check if the words are correctly aligned
        if len(gloss_words) != len(utterance_words):
            return gloss_tier
        else:
            new_gloss_words = []
            for gloss_word, actual_word in zip(gloss_words, utterance_words):
                if 'Eng' in gloss_word:
                    new_gloss_words.append(actual_word)
                else:
                    new_gloss_words.append(gloss_word)
            return ' '.join(new_gloss_words)

    @classmethod
    def utterance_cross_clean(
            cls, raw_utt, actual_utt, target_utt,
            seg_tier, gloss_tier, pos_tier):
        gloss_tier = cls.replace_eng(gloss_tier, actual_utt)
        return actual_utt, target_utt, seg_tier, gloss_tier, pos_tier

    # ---------- morpheme word cleaning ----------

    @staticmethod
    def replace_percentages(word):
        """Replace words consisting of percentages.

        '%%%' stand for untranscribed words. They are replaced by '???'.
        """
        if word == '%%%':
            return '???'
        else:
            return word

    @staticmethod
    def replace_hashtag(morph_element):
        """Replace words and morphemes consisting of a hashtag.

        '#' stands for unglossed words and morphemes. It is replaced by
        '???'
        """
        if morph_element == '#':
            return '???'
        else:
            return morph_element

    @staticmethod
    def handle_question_mark(morph_element):
        """Handle question marks in words and morphemes.

        '?' stands for unclear meanings of a morpheme and word. If it only
        consists of '?', it is replaced by '???'. If there is a form followed
        by '?', it is removed.
        """
        if morph_element == '?':
            return '???'

        return morph_element.replace('?', '')

    @classmethod
    def replace_star(cls, morph_element):
        """Replace words or morphemes consisting of a star.

        The star marks an element that does not correspond to an element on
        another morphology tier. It is replaced by a '???'.
        """
        if morph_element == '*':
            return '???'
        else:
            return morph_element

    @classmethod
    def clean_morpheme_word(cls, morpheme_word):
        for cleaning_method in [
                cls.replace_percentages, cls.replace_hashtag,
                cls.handle_question_mark, cls.replace_star]:
            morpheme_word = cleaning_method(morpheme_word)

        return morpheme_word

    # ---------- morpheme cleaning ----------

    @staticmethod
    def remove_parentheses(segment):
        return segment.lstrip('(').rstrip(')')

    @classmethod
    def clean_segment(cls, segment):
        for cleaning_method in [
                cls.remove_parentheses,
                cls.replace_hashtag,
                cls.handle_question_mark,
                cls.replace_star]:
            segment = cleaning_method(segment)

        return segment

    @classmethod
    def clean_gloss(cls, gloss):
        return CreeGlossMapper.map(gloss)

    @classmethod
    def clean_pos(cls, pos):
        return CreePOSMapper.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return CreePOSMapper.map(pos_ud, ud=True)
